"""2D Polygon Magnet class


"""
import numpy as _np
from ._magnet2D import Magnet2D
from ..utils.global_const import PI, MAG_TOL


def _sheet_field(x, y, h, Kr=1):
    """Calculates the magnetic field due to an infinite current sheet

    Args:
        x (ndarray): x-coordinates
        y (ndarray): y-coordinates
        h (float): sheet length
        Kr (float, optional): Sheet current Density as T. Defaults to 1.

    Returns:
        tuple: Bx (ndarray), By (ndarray) magnetic field vector
    """
    x = _np.asarray(x)
    y = _np.asarray(y)
    prefac = Kr / 4 / PI
    Bx = prefac * _np.log((x ** 2 + (y - h) ** 2) / (x ** 2 + (y + h) ** 2))
    By = 2 * prefac * _np.arctan2(2 * h * x, x ** 2 + y ** 2 - h ** 2)
    return Bx, By


class Polygon(object):
    """Polygon class for generating list of vertices"""

    def __init__(self, **kwargs):
        vertices = kwargs.pop("vertices", None)
        center = kwargs.pop("center", None)
        if vertices is not None:
            if type(vertices) is _np.ndarray:
                if center is not None:
                    vertices += center
                self.vertices = vertices.tolist()
            else:
                self.vertices = vertices

            if center is not None:
                self.center = center
            else:
                self.set_center()
        else:
            self.vertices = []
            self.center = _np.NaN

    def append(self, vertex):
        """Appends vertex to list of vertices

        Args:
            vertex (list): list of vertices
        """
        if len(vertex) != 2:
            print("Error")
        if type(vertex) == tuple:
            self.vertices.append(vertex)
        elif len(vertex) == 2:
            self.vertices.append(tuple(vertex))
        self.set_center()

    def num_vertices(self):
        """Gets number of vertices

        Returns:
            int: number of vertices
        """
        return len(self.vertices)

    def set_center(self):
        """Sets center of polygon to be centroid"""
        self.center = _np.mean(_np.asarray(self.vertices), axis=0)

    @staticmethod
    def gen_polygon(N=6, center=(0.0, 0.0), alpha=0.0, **kwargs):
        """Generates regular polygon. One of apothem, side length or radius must be defined.

        Args:
            N (int, optional): Number of sides. Defaults to 6.
            center (tuple, optional): Polygon center. Defaults to (0.0, 0.0).
            alpha (float, optional): Orientration with respect to x-axis. Defaults to 0.0.

        Raises:
            Exception: N must be > 2

        Returns:
            ndarray: polygon vertices
        """
        N = int(N)

        if N < 3:
            raise Exception("Error, N must be > 2.")

        apothem = kwargs.pop("apothem", None)
        length = kwargs.pop("length", None)
        radius = kwargs.pop("radius", None)

        radius = Polygon.check_radius(N, apothem, length, radius)

        k = _np.arange(0, N, 1)
        xc = center[0]
        yc = center[1]

        def f(N):
            if N % 2 == 0:
                return PI / N + _np.deg2rad(alpha)
            else:
                return PI / N + PI + _np.deg2rad(alpha)

        xv = xc + radius * _np.sin(2 * PI * k / N + f(N))
        yv = yc + radius * _np.cos(2 * PI * k / N + f(N))
        poly_verts = _np.vstack((xv, yv)).T.tolist()

        return poly_verts

    @staticmethod
    def check_radius(N, apothem, length, radius):
        """Checks which of apothem, side length, or radius has been passed as kwargs
        to `gen_polygon()`. Order of precendence is apothem, length, radius.

        Args:
            N (int): Number of sides
            apothem (float): polygon apothem
            length (float): side length
            radius (float): outcircle radius

        Raises:
            Exception: One of apothem, length, or raduis must be defined

        Returns:
            float: returns radius
        """
        if apothem is not None:
            return apothem / _np.around(_np.cos(PI / N), 4)
        elif length is not None:
            return length / _np.around(2 * _np.sin(PI / N), 4)
        elif radius is not None:
            return radius
        else:
            raise Exception("Error, one of apothem, length, or radius must be defined.")


class LineUtils(object):
    """Utility class consisting of rountines for 2D line elements"""

    @staticmethod
    def unit_norm(vertex_1, vertex_2, clockwise=True):
        """Get unit normal to vertex

        Args:
            vertex_1 (ndarray): vertex 1
            vertex_2 (ndarray): vertex 2
            clockwise (bool, optional): Clockwise orientation of points. Defaults to True.

        Returns:
            tuple: normal vector (ndarray), length i.e. distance between vertices (float)
        """

        dx = vertex_1[0] - vertex_2[0]
        dy = vertex_1[1] - vertex_2[1]

        # Clockwise winding of points:
        if clockwise:
            norm = _np.array([dy, -dx])
        else:
            norm = _np.array([-dy, dx])
        length = _np.linalg.norm(norm)
        norm = norm / length
        return norm, length

    @staticmethod
    def line_center(vertex_1, vertex_2):
        """Gets midpoint of two vertices

        Args:
            vertex_1 (ndarray): vertex 1
            vertex_2 (ndarray): vertex 2

        Returns:
            ndarray: midpoint
        """
        xc = (vertex_1[0] + vertex_2[0]) / 2
        yc = (vertex_1[1] + vertex_2[1]) / 2

        return _np.array([xc, yc])

    @staticmethod
    def signed_area(polygon):
        """Calculates signed area of a polygon

        Args:
            polygon (Polygon): Polygon instance

        Returns:
            float: signed area
        """
        j = 1
        NP = polygon.num_vertices()
        area = 0
        norm = _np.empty([NP, 2])
        center = _np.empty([NP, 2])
        beta = _np.empty(NP)  # angle w.r.t. y axis
        length = _np.empty(NP)

        for i in range(NP):
            j = j % NP
            area += (polygon.vertices[j][0] - polygon.vertices[i][0]) * (
                polygon.vertices[j][1] + polygon.vertices[i][1]
            )
            norm[i, :], length[i] = LineUtils.unit_norm(
                polygon.vertices[i], polygon.vertices[j]
            )
            center[i, :] = LineUtils.line_center(
                polygon.vertices[i], polygon.vertices[j]
            )
            j += 1

        # check winding order of polygon, area < 0 for clockwise ordering of points
        if area < 0:
            norm *= -1
        beta[:] = _np.rad2deg(_np.arctan2(norm[:, 1], norm[:, 0]))

        return area / 2.0, norm, beta, length, center


class Line(object):
    """Line Class for storing properties of a sheet manget"""

    def __init__(self, length, center, beta, K):
        """Init Method

        Args:
            length (float): side length
            center (ndarray): magnet center (x, y)
            beta (float): Orientation w.r.t. z-axis in degrees
            K (float): Sheet current density in tesla
        """
        self.length = length
        self.center = center
        self.beta = beta
        self.beta_rad = _np.deg2rad(beta)
        self.xc = center[0]
        self.yc = center[1]
        self.K = K
        self.tol = MAG_TOL

    def __str__(self):
        str = (
            f"K: {self.K} (T)\n"
            + f"Length: {self.length} (m)\n"
            + f"Center {self.center} (m)\n"
            + f"Orientation: {self.beta}\n"
        )
        return str

    def __repr__(self):
        str = (
            f"K: {self.K} (T)\n"
            + f"Length: {self.length}\n"
            + f"Center {self.center}\n"
            + f"Orientation: {self.beta}\n"
        )
        return str

    def get_center(self):
        """Returns line center

        Returns:
            ndarray: center (x,y)
        """

        return self.center

    def get_field(self, x, y):
        """Calculates the magnetic field due to a sheet magnet
        First a transformation into the local coordinates is made, the field calculated
        and then the magnetic field it rotated out to the global coordinates

        Args:
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates

        Returns:
            tuple: Bx (ndarray), By (ndarray) magnetic field vector
        """
        from ..utils._routines2D import rotate_points_2D, _get_field_array_shape2

        array_shape = _get_field_array_shape2(x, y)
        Bx, By = _np.zeros(array_shape), _np.zeros(array_shape)
        if _np.fabs(self.beta_rad) > self.tol:
            xt, yt = rotate_points_2D(x - self.xc, y - self.yc, 2 * PI - self.beta_rad)
            Btx, Bty = _sheet_field(xt, yt, self.length / 2, self.K)
            Bx, By = rotate_points_2D(Btx, Bty, self.beta_rad)

        else:
            Bx, By = _sheet_field(x - self.xc, y - self.yc, self.length / 2, self.K)
        return Bx, By


class PolyMagnet(Magnet2D):
    """2D Magnet Polygon class."""

    mag_type = "PolyMagnet"

    def __init__(self, Jr, **kwargs) -> None:
        """Init method

        NOTE:
            * When creating a regular polygon, one of apothem, radius, or length must be defined as a kwarg or an exception will be raised.
            * When creating a regular polygon, the number of sides `num_sides` must be at least 3 or an exception will be raised.
            * When creating a custom polygon at least one vertex pair must be defined with `vertices` or an exception will be raised.

        Args:
            Jr (float): signed magnitude of remnant magnetisation

        Kwargs:
            alpha (float): Not used
            theta (float): Orientation of magnet w.r.t x-axis of magnet
            phi (float): Orientation of magnetisation w.r.t x-axis of magnet in degrees. Defaults to 90.0.
            center (ndarray): magnet center (x, y). Defaults to (0.0, 0.0)
            length (float): side length if creating a regular polygon
            apothem (float): apothem (incircle radius) if creating a regular polygon
            radius (float): radius (circumcircle radius) if  creating a regular polygon
            num_sides (int): number of sides of a regular polygon. Defaults to 6.
            custom_polygon (bool): Flag to define a custom polygon. Defaults to False.
            vertices (ndarray, list): List of custom vertices. Defaults to None.

        Raises:
            Exception: If creating a custom polygon, `vertices` must not be None.
        """
        from ..utils._routines2D import rotate_points_2D

        super().__init__(Jr, **kwargs)

        # Magnet rotation w.r.t. x-axis
        self.alpha = kwargs.pop("alpha", 0.0)
        self.alpha_radians = _np.deg2rad(self.alpha)

        self.theta = kwargs.pop("theta", 0.0)
        self.theta_radians = _np.deg2rad(self.theta)

        self.phi = kwargs.pop("phi", 90.0)
        self.phi_rad = _np.deg2rad(self.phi)

        self.Jx = _np.around(Jr * _np.cos(self.phi_rad), decimals=6)
        self.Jy = _np.around(Jr * _np.sin(self.phi_rad), decimals=6)
        self.tol = MAG_TOL
        self.area = None

        self.custom_polygon = kwargs.pop("custom_polygon", False)

        self.center = kwargs.pop("center", _np.array([0.0, 0.0, 0.0]))
        self.center = _np.asarray(self.center)

        if self.custom_polygon:
            vertices = kwargs.pop("vertices", None)
            if vertices is None:
                raise Exception("Error, no vertices were defined.")

            vertices = _np.atleast_2d(vertices)

            x_rot, y_rot = rotate_points_2D(
                vertices[:, 0],
                vertices[:, 1],
                self.theta_radians,  # + self.alpha_radians,
            )
            vertices = _np.stack([x_rot, y_rot]).T + self.center
            self.polygon = Polygon(vertices=vertices.tolist())
        else:
            self.length = kwargs.pop("length", None)
            self.apothem = kwargs.pop("apothem", None)
            self.radius = kwargs.pop("radius", None)
            self.num_sides = kwargs.pop("num_sides", 6)

            self.radius = Polygon.check_radius(
                self.num_sides,
                self.apothem,
                self.length,
                self.radius,
            )
            # Generate Polygon
            self.polygon = Polygon(
                vertices=Polygon.gen_polygon(
                    self.num_sides,
                    self.center,
                    self.theta,  # + self.alpha,
                    length=self.length,
                    apothem=self.apothem,
                    radius=self.radius,
                ),
                center=self.center,
            )

    def get_center(self):
        """Returns magnet centre

        Returns:
            center (ndarray): numpy array [xc, yc]
        """
        return self.center

    def get_orientation(self):
        """Returns magnet orientation, `alpha` in degrees

        Returns:
            float: alpha, rotation angle w.r.t x-axis.
        """

        return self.alpha

    def _gen_sheet_magnets(self):
        """Generates orientation, size, and centre of sheet magnets for a given
        polygon

        Returns:
            tuple: beta (ndarray), length (ndarray), centre (ndarray), K (ndarray) - sheet current density in tesla.
        """
        area, norms, beta, length, center = LineUtils.signed_area(self.polygon)
        K = self.Jx * norms[:, 1] - self.Jy * norms[:, 0]
        self.area = area
        return beta, length, center, K

    def get_field(self, x, y):
        """Calculates the magnetic field of a polygon due to each face

        Args:
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates

        Returns:
            tuple: Bx (ndarray), By (ndarray) magnetic field vector
        """
        from ..utils._routines2D import rotate_points_2D, _get_field_array_shape2

        array_shape = _get_field_array_shape2(x, y)
        Bx, By = _np.zeros(array_shape), _np.zeros(array_shape)
        beta, length, center, K = self._gen_sheet_magnets()

        if _np.fabs(self.alpha_radians) > self.tol:
            pass
            print("Arbitrary rotation with alpha not yet implemented!!")

            # FIXME: rotate centres
            # xt, yt = rotate_points_2D(x - self.xc, y - self.yc, self.alpha_radians)
            # beta += self.alpha
            # xc_rot, yc_rot = rotate_points_2D(
            #     center[:, 0] - self.xc,
            #     center[:, 1] - self.yc,
            #     self.alpha_radians,
            # )
            # center[:, 0] = xc_rot
            # center[:, 1] = yc_rot
            #
            #
            # for i in range(len(K)):
            #     sheet = Line(length[i], center[i], beta[i], K[i])
            #     Btx, Bty = sheet.get_field(xt, yt)
            #     Btx, Bty = rotate_points_2D(Btx, Bty, 2 * PI - self.alpha_radians)
            #     Bx += Btx
            #     By += Bty

        for i in range(len(K)):
            sheet = Line(length[i], center[i], beta[i], K[i])
            Btx, Bty = sheet.get_field(x, y)
            Bx += Btx
            By += Bty
        return Bx, By



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""pymagnets.utils._point_structs

Private module consiting of point classes and their methods.
"""

__all__ = ["Point2", "Point3"]

import numpy as _np


class Point2(object):
    """2D point class

    Note that multiplication of two points is done elementwise, dot product is
    a separate method.
    """

    def __init__(self, x, y):
        """Init method

        Args:
            x (ndarray): x coordinates
            y (ndarray): y coordinates
        """
        self.x = x
        self.y = y

    def __repr__(self) -> str:
        return f"({self.x}, {self.y})"

    def __str__(self) -> str:
        return f"({self.x}, {self.y})"

    def __add__(self, other):
        return Point2(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        return Point2(self.x - other.x, self.y - other.y)

    def __mul__(self, other):
        return Point2(self.x * other.x, self.y * other.y)

    def __div__(self, other):
        x = (self.x / other.x) if other.x != 0 else 0
        y = (self.y / other.y) if other.y != 0 else 0
        return Point2(x, y)

    def __lt__(self, other):
        self_mag = (self.x ** 2) + (self.y ** 2)
        other_mag = (other.x ** 2) + (other.y ** 2)
        return self_mag < other_mag

    def __le__(self, other):
        self_mag = (self.x ** 2) + (self.y ** 2)
        other_mag = (other.x ** 2) + (other.y ** 2)
        return self_mag <= other_mag

    def __gt__(self, other):
        self_mag = (self.x ** 2) + (self.y ** 2)
        other_mag = (other.x ** 2) + (other.y ** 2)
        return self_mag > other_mag

    def __ge__(self, other):
        self_mag = (self.x ** 2) + (self.y ** 2)
        other_mag = (other.x ** 2) + (other.y ** 2)
        return self_mag >= other_mag

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    def __ne__(self, other):
        return self.x != other.x or self.y != other.y

    def distance_to(self, point):
        """Calculates distance to a point

        Args:
            point (Point2): Target point

        Returns:
            float: distance
        """
        return _np.sqrt(_np.power(point.x - self.x, 2) + _np.power(point.y - self.y, 2))

    def distance_to_origin(self):
        """Calculates distance to a origin

        Returns:
            float: distance
        """
        return self._norm()

    def _norm(self):
        """Norm of Point

        Returns:
            float: norm
        """
        return _np.linalg.norm([self.x, self.y], axis=0)


#     def move(self, point):
#         """Translates point by
#
#         Args:
#             point (Point2): point
#         """
#         try:
#             if type(point) is tuple:
#                 self.x += point[0]
#                 self.y += point[1]
#             else:
#                 self + point
#         except ValueError as e:
#             print(f"Invalid input {0}, should be a 2-element tuple or Point2 object", e)


class Point3(Point2):
    """3D point class

    Note that multiplication of two points is done elementwise, dot product is
    a separate method.

    """

    def __init__(self, x, y, z):
        super().__init__(x, y)
        self.z = z

    def __repr__(self) -> str:
        return f"({self.x}, {self.y}, {self.z})"

    def __str__(self) -> str:
        return f"({self.x}, {self.y}, {self.z})"

    def __add__(self, other):
        return Point3(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other):
        return Point3(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, other):
        return Point3(self.x * other.x, self.y * other.y, self.z * other.z)

    def __div__(self, other):
        x = (self.x / other.x) if other.x != 0 else 0
        y = (self.y / other.y) if other.y != 0 else 0
        z = (self.z / other.z) if other.z != 0 else 0
        return Point3(x, y, z)

    def __lt__(self, other):
        self_mag = (self.x ** 2) + (self.y ** 2)
        other_mag = (other.x ** 2) + (other.y ** 2)
        return self_mag < other_mag

    def __le__(self, other):
        self_mag = (self.x ** 2) + (self.y ** 2)
        other_mag = (other.x ** 2) + (other.y ** 2)
        return self_mag <= other_mag

    def __gt__(self, other):
        self_mag = (self.x ** 2) + (self.y ** 2)
        other_mag = (other.x ** 2) + (other.y ** 2)
        return self_mag > other_mag

    def __ge__(self, other):
        self_mag = (self.x ** 2) + (self.y ** 2)
        other_mag = (other.x ** 2) + (other.y ** 2)
        return self_mag >= other_mag

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    def __ne__(self, other):
        return self.x != other.x or self.y != other.y

    def distance_to(self, point):
        return _np.sqrt(
            _np.power(point.x - self.x, 2)
            + _np.power(point.y - self.y, 2)
            + _np.power(point.z - self.z, 2)
        )

    def _norm(self):
        return _np.linalg.norm([self.x, self.y, self.z], axis=0)

    def distance_to_origin(self):
        return self._norm()



from ..magnets import Magnet3D
from ..utils.global_const import FP_CUTOFF, MU0, ALIGN_CUTOFF
from ..utils._conversions import get_unit_value_meter
from ..utils._routines3D import _allocate_field_array3
from ..utils._vector_structs import Point_Array3
import numpy as _np
from numba import njit, guvectorize

__all__ = [
    "get_centroid",
    "triangle_area",
    "get_area_triangles",
    "get_midpoints",
    "divide_triangle_centroid",
    "divide_triangle_regular",
]


@njit
def get_centroid(triangle):
    """Gets the centroid of a triangle

    Args:
        triangle (ndarray): (3,3) array of vertices of a triangle in 3D

    Returns:
        ndarray: (3,) array of centroid coordinates
    """
    result = _np.zeros(3)
    for i in range(3):
        result[i] = (triangle[0, i] + triangle[1, i] + triangle[2, i]) / 3.0
    return result


# @guvectorize(['void(f8[:,:], f8[:])'],
#               '(x, x)->(x)')
# def get_centroid(triangle, result):
#     for i in range(3):
#         result[i] = (triangle[0,i] + triangle[1,i] + triangle[2,i])/3.0


@njit
def triangle_area(triangle):
    """Gets the area of a triangle. Computes the cross product area.

    Args:
        triangle (ndarray): (3,3) array of triangle vertices in 3D

    Returns:
        float: area
    """
    return _np.linalg.norm(
        _np.cross((triangle[1] - triangle[0]), (triangle[2] - triangle[0])) / 2
    )


# @guvectorize(["void(f8[:,:, :], f8[:])"], "(x, y, y)->(x)")
@njit
def get_area_triangles(triangles, area):
    """Computes the area for an array of triangles

    Args:
        triangles (ndarray): (N,3,3) array of N triangles in 3D
        area (ndarray): (N,) array of areas
    """
    for i in range(len(triangles)):
        area[i] = _np.linalg.norm(
            _np.cross(
                (triangles[i][1] - triangles[i][0]), (triangles[i][2] - triangles[i][0])
            )
            / 2
        )


@njit
def _divide_triangle_centroid(triangle, depth=1, memo=_np.array(())):
    """Recursively divides a triangle into 3 using the centroid

    Args:
        triangle (ndarray): (3,3) array of triangle vertices in 3D
        depth (int, optional): Number of recursions to do. Defaults to 1.
        memo (ndarray, optional): Tracked array of generated vertices. Defaults to _np.array(()).

    Returns:
        ndarray: N*3*3 array of vertices. where N = 3**depth
    """

    centroid = _np.array([0.0, 0.0, 0.0])
    trig_array = _np.zeros((3, 3, 3))
    centroid = get_centroid(triangle)

    for i in range(3):
        centroid[i] = (triangle[0, i] + triangle[1, i] + triangle[2, i]) / 3
        trig_array[i, :] = triangle[:]

    trig_array[0, 2, :] = centroid
    trig_array[1, 0, :] = centroid
    trig_array[2, 1, :] = centroid
    depth -= 1

    if depth < 1:
        memo = _np.append(memo, trig_array)
    else:
        memo = _divide_triangle_centroid(trig_array[0, ...], depth=depth, memo=memo)
        memo = _divide_triangle_centroid(trig_array[1, ...], depth=depth, memo=memo)
        memo = _divide_triangle_centroid(trig_array[2, ...], depth=depth, memo=memo)

    return memo


@njit
def get_midpoints(triangle):
    """Get midpoints of the faces of a triangle

    Args:
        triangle (ndarray): (3,3) array of triangle vertices in 3D

    Returns:
        ndarray: (3,3) array of the three midpoints
    """
    midpoints = _np.zeros_like(triangle)
    for i in range(3):
        midpoints[i, ...] = (triangle[i] + triangle[(i + 1) % 3]) / 2
    return midpoints


@njit
def _divide_triangle_regular(triangle, depth=1, memo=_np.array(())):
    """Recursively divides a triangle into 4 using the midpoint of each face.

    Args:
        triangle (ndarray): (3,3) array of triangle vertices in 3D
        depth (int, optional): Number of recursions to perform. Defaults to 1.
        memo (ndarray, optional): Tracked array of generated vertices. Defaults to _np.array(()).

    Returns:
        ndarray: (N*3*3,) array of vertices, where N = 4**depth
    """
    midpoints = get_midpoints(triangle)
    trig_array = _np.zeros((4, 3, 3))
    trig_array[0, 0, :] = triangle[0]
    trig_array[0, 1, :] = midpoints[0]
    trig_array[0, 2, :] = midpoints[2]

    trig_array[1, 0, :] = midpoints[0]
    trig_array[1, 1, :] = midpoints[1]
    trig_array[1, 2, :] = midpoints[2]

    trig_array[2, 0, :] = midpoints[2]
    trig_array[2, 1, :] = midpoints[1]
    trig_array[2, 2, :] = triangle[2]

    trig_array[3, 0, :] = midpoints[0]
    trig_array[3, 1, :] = triangle[1]
    trig_array[3, 2, :] = midpoints[1]
    depth -= 1

    if depth < 1:
        memo = _np.append(memo, trig_array)
    else:
        memo = _divide_triangle_regular(trig_array[0, ...], depth=depth, memo=memo)
        memo = _divide_triangle_regular(trig_array[1, ...], depth=depth, memo=memo)
        memo = _divide_triangle_regular(trig_array[2, ...], depth=depth, memo=memo)
        memo = _divide_triangle_regular(trig_array[3, ...], depth=depth, memo=memo)

    return memo


def divide_triangle_centroid(triangle, depth=1):
    """Recursively divides a triangle into 3 using the centroid.
    A wrapper over `_divide_triangle_centroid` to  reshape the output from
    (N*3*3,) to (N,3,3)

    Args:
        triangle (ndarray): (3,3) array of triangle vertices in 3D
        depth (int, optional): Number of recursions to perform. Defaults to 1.

    Returns:
        ndarray: (N,3,3) array of vertices, where N = 3**depth
    """
    mesh = _divide_triangle_centroid(triangle, depth=depth)
    mesh = mesh.reshape((mesh.shape[0] // 9, 3, 3))
    return mesh


def divide_triangle_regular(triangle, depth=1):
    """Recursively divides a triangle into 4 using the midpoint of each face.
    A wrapper over `_divide_triangle_regular` to  reshape the output from
    (N*3*3,) to (N,3,3)

    Args:
        triangle (ndarray): (3,3) array of triangle vertices in 3D
        depth (int, optional): Number of recursions to perform. Defaults to 1.

    Returns:
        ndarray: (N,3,3) array of vertices, where N = 4**depth
    """
    mesh = _divide_triangle_regular(triangle, depth=depth)
    mesh = mesh.reshape((mesh.shape[0] // 9, 3, 3))
    return mesh


def _calc_field_simplex(active_magnet, points):
    """Calculates the total force and torque acting on one simplex of a magnet due
    to all other instantiated magnets

    Args:
        active_magnet (Magnet3D): Target magnet
        points (Point_Array3): Grid of points (x,y,z)

    Returns:
        tuple: total_field (float), total_torque (float)
    """
    field = _allocate_field_array3(points.x, points.y, points.z)
    xc, yc, zc = active_magnet.centroid

    for magnet in Magnet3D.instances:
        if magnet is not active_magnet:
            Bx, By, Bz = magnet.get_field(points.x, points.y, points.z)
            field.x += Bx.reshape(field.x.shape)
            field.y += By.reshape(field.y.shape)
            field.z += Bz.reshape(field.z.shape)

    field.x[~_np.isfinite(field.x)] = 0.0
    field.y[~_np.isfinite(field.y)] = 0.0
    field.z[~_np.isfinite(field.z)] = 0.0

    total_field = _np.array((_np.sum(field.x), _np.sum(field.y), _np.sum(field.z)))

    torques = _np.cross(
        _np.array(
            [points.x.ravel() - xc, points.y.ravel() - yc, points.z.ravel() - zc]
        ).T,
        _np.array([field.x.ravel(), field.y.ravel(), field.z.ravel()]).T,
    )
    total_torque = _np.sum(torques, axis=0)

    return total_field, total_torque


def calc_force_mesh(active_magnet, depth=3, unit="mm"):
    force = _np.zeros(3)
    torque = _np.zeros(3)

    for i in range(len(active_magnet.mesh_vectors)):
        if _np.fabs(active_magnet.Jnorm[i] / active_magnet.Jr) > ALIGN_CUTOFF:
            triangle = active_magnet.mesh_vectors[i]
            mesh = divide_triangle_regular(triangle, depth=depth)
            centroids = _np.mean(mesh, axis=1)
            num_sub_triangles = len(centroids)

            points = Point_Array3(
                centroids[:, 0], centroids[:, 1], centroids[:, 2], unit=unit
            )
            local_field, local_torque = _calc_field_simplex(active_magnet, points)
            area = triangle_area(triangle)
            force += local_field * -active_magnet.Jnorm[i] * area / num_sub_triangles
            torque += local_torque * -active_magnet.Jnorm[i] * area / num_sub_triangles

    scaling_factor = get_unit_value_meter(points.get_unit())
    force /= MU0 / scaling_factor / scaling_factor
    torque /= MU0 / scaling_factor / scaling_factor / scaling_factor

    return force, torque


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""pymagnet
User friendly magnetic field calculations

Top level module for exposing the public API of usable modules
"""
from . import plots
from . import magnets
from . import utils
from . import forces
from .magnets._magnet_base import reset_magnets, list_magnets
from .utils._routines2D import grid2D, get_field_2D
from .utils._routines3D import grid3D, slice3D, get_field_3D

__all__ = ["magnets", "plots", "utils"]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""1D High symmetry methods

This private module implements magnetic field calculations in z along the
symmetry centre of a cylindrical or cuboidal magnet.

"""
import numpy as _np
from ..utils.global_const import PI
from ..utils._vector_structs import Field1

__all__ = ["magnetic_field_prism_1D", "magnetic_field_cylinder_1D"]


def magnetic_field_prism_1D(magnet, z):
    """Calculates the magnetic field z-component due to a cuboid along its
    axial symmetry center.

    Args:
        magnet (Magnet3D): magnet object, one of Prism, or Cube
        z (ndarray): Array of points along the symmetry axis

    Returns:
        Field1: z-component of the magnetic field and associated unit ('T')
    """
    from ._magnet3D import Prism

    if issubclass(magnet.__class__, Prism):
        a = magnet.a
        b = magnet.b
        c = magnet.c
        Jr = magnet.Jr

        z_local = _np.asarray(z) - c - magnet.center[2]

        ab = a * b
        a_sq = _np.power(a, 2)
        b_sq = _np.power(b, 2)
        z_sq = _np.power(z_local, 2)
        zc = z_local + 2 * c
        zc_sq = _np.power(zc, 2)

        Bz = _np.arctan2(zc * _np.sqrt(a_sq + b_sq + zc_sq), ab) - _np.arctan2(
            z_local * _np.sqrt(a_sq + b_sq + z_sq), ab
        )
        Bz *= Jr / PI
        field = Field1(Bz)
        return field
    else:
        print(f"Error, the magnet should be a 3D magnet not {magnet.__class__}")
        return None


def magnetic_field_cylinder_1D(magnet, z):
    """Calculates the magnetic field z-component due to a cuboid along its
    axial symmetry center.

    Args:
        magnet (Magnet3D): magnet object, Cylinder
        z (ndarray): Array of points along the symmetry axis

    Returns:
        Field1: z-component of the magnetic field and associated unit ('T')
    """
    from ._magnet3D import Cylinder

    if issubclass(magnet.__class__, Cylinder):
        L = magnet.length
        R = magnet.radius
        Jr = magnet.Jr

        z_local = _np.asarray(z) - magnet.length / 2 - magnet.center[2]

        zL = z_local + L
        R_sq = _np.power(R, 2)
        z_sq = _np.power(z_local, 2)
        zL_sq = _np.power(zL, 2)

        Bz = (zL / _np.sqrt(zL_sq + R_sq)) - (z_local / _np.sqrt(z_sq + R_sq))
        Bz *= Jr / 2
        data = Field1(Bz)
        return data

    else:
        print(f"Error, the magnet should be a 3D magnet not {magnet.__class__}")
        return None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""pymagnets.forces

Forcs and Torques

This module imports the classes and functions in the private modules to create a public API.

"""
from ._cylinder_force import *
from ._mesh_force import *
from ._prism_force import *
from ._sphere_force import *


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""2D plotting routines using Plotly as the backend: NOT IMPLEMENTED YET

This module contains all functions needed to plot lines and contours for 2D
magnetic sources.

"""
import plotly.graph_objects as _go
import plotly.figure_factory as _ff
import numpy as _np
from ..utils._conversions import get_unit_value_meter


def _plotly_vector_plot2(x, y, Field, NQ, scale_x, scale_y, vector_color):
    """Draws quiver plot of field vectors

    Args:
        x (float/array): x coordinates
        y (float/array): y coordinates
        Field (Vector2): Field structure
        NQ (int): Plot every NQth vector
        scale_x (float): unit scaling
        scale_y (float): unit scaling
        vector_color (string): quiver color
    """
    plot_object = []
    NPx, NPy = x.shape
    if NQ != 0:
        NSx, NSy = NPx // NQ, NPy // NQ
        with _np.errstate(divide="ignore", invalid="ignore"):
            plot_object = _ff.create_quiver(
                x[::NSx, ::NSy] / scale_x,
                y[::NSx, ::NSy] / scale_y,
                Field.x[::NSx, ::NSy] / Field.n[::NSx, ::NSy],
                Field.y[::NSx, ::NSy] / Field.n[::NSx, ::NSy],
                scale=1.5,
                arrow_scale=0.5,  # name='quiver',
                line_width=1,
                line=dict(color="white"),
            )

    return plot_object


def plotly_2D_contour(x, y, Field, **kwargs):
    """Contour Plot of field

    Args:
        x ([array]): [assumes in µm]
        y ([array]): [assumes in nm]
        Field ([field vector_2D]): [field vector object]
    """

    scale_x = kwargs.pop("scale_x", 1)
    scale_y = kwargs.pop("scale_y", 1)
    scale_cb = kwargs.pop("scale_cb", 1)

    colorscale = kwargs.pop("colorscale", "viridis")
    cmin = kwargs.pop("cmin", 0.0)
    cmax = kwargs.pop("cmax", round(_np.nanmean(Field.n[:]) * 2, 1))
    cstep = kwargs.pop("cstep", 0.1)

    show_magnets = kwargs.pop("show_magnets", True)
    title = kwargs.pop("title", None)
    xlab = kwargs.pop("xlab", "x (mm)")
    ylab = kwargs.pop("ylab", "y (mm)")
    clab = kwargs.pop("clab", "B (T)")

    # field_component = kwargs.pop("field_component", "n")

    # plot_type = kwargs.pop("plot_type", "contour")

    plot_objects = []
    # cache = {}

    plot_objects.append(
        _plotly_contour2(
            x,
            y,
            Field.n,
            cmin=cmin,
            cmax=cmax,
            cstep=cstep,
            colorscale=colorscale,
        )
    )

    plot_objects.append(_plotly_draw_circle(p1=(-10, -10), p2=(10, 10)))
    plot_objects.append(_plotly_draw_arrow(head=(6, -6), tail=(-6, 6)))

    fig = _go.Figure(data=plot_objects)

    fig.update_layout(
        title=title,
        xaxis_title=xlab,
        yaxis_title=ylab,
    )
    fig.update_yaxes(
        scaleanchor="x",
        scaleratio=1,
    )

    fig.update_xaxes(range=[x.min() * 1, x.max() * 1])
    fig.update_yaxes(range=[y.min() * 1, y.max() * 1])
    # fig.update_yaxes(automargin=True)

    fig.update_layout(
        autosize=False,
        width=500,
        height=500,
        margin=dict(l=50, r=50, b=50, t=100, pad=0),
    )

    fig.show()


def _plotly_contour2(x, y, z, **kwargs):
    colorscale = kwargs.pop("colorscale", "viridis")
    cmin = kwargs.pop("cmin", 0.0)
    cmax = kwargs.pop("cmax", round(z.mean() * 2, 1))
    cstep = kwargs.pop("cstep", 0.1)

    plot_object = [
        _go.Contour(
            z=z,
            x=x,  # horizontal axis
            y=y,  # vertical axis
            colorscale=colorscale,
            connectgaps=True,
            line_smoothing=0.85,
            zmin=cmin,
            zmax=cmax,
            contours=dict(
                # coloring="heatmap",
                showlabels=True,  # show labels on contours
                start=cmin,
                end=cmax,
                size=cstep,
                labelfont=dict(  # label font properties
                    size=12,
                    color="white",
                ),
            ),
            colorbar=dict(
                title="|B| (T)",
                titleside="right",
                titlefont=dict(size=14, family="Arial, sans-serif"),
            ),
        )
    ]
    return plot_object


def _plotly_draw_circle(p1=(-0.5, -0.5), p2=(0.5, 0.5), **kwargs):
    fillcolor = kwargs.pop("fillcolor", "white")
    line_color = kwargs.pop("line_color", "black")
    return dict(
        type="circle",
        xref="x",
        yref="y",
        fillcolor=fillcolor,
        x0=p1[0],
        y0=p1[1],
        x1=p2[0],
        y1=p2[1],
        line_color=line_color,
    )


def _plotly_draw_arrow(head=(-0.5, -0.5), tail=(0.5, 0.5), **kwargs):

    arrowcolor = kwargs.pop("arrowcolor", "black")
    return dict(
        x=head[0],  # arrows' head
        y=head[1],  # arrows' head
        ax=tail[0],  # arrows' tail
        ay=tail[1],  # arrows' tail
        xref="x",
        yref="y",
        axref="x",
        ayref="y",
        text="",  # if you want only the arrow
        showarrow=True,
        arrowhead=3,
        arrowsize=1,
        arrowwidth=5,
        arrowcolor=arrowcolor,
    )



__all__ = ["calc_force_prism"]

import numpy as _np
from ..utils.global_const import FP_CUTOFF, MU0
from ..utils._conversions import get_unit_value_meter
from ..utils._vector_structs import Point_Array3
from ..utils._routines3D import _allocate_field_array3
from ..utils._quaternion import Quaternion
from ..magnets import Magnet3D


def _get_ranges_prism(active_magnet):
    """Gets min and max x,y,z values for magnet faces.

    This is needed to generate a grid of points on each surface for calculating
    the force and torque on the magnet due to the other magnets.

    Args:
        active_magnet (Prism): Target magnet

    Returns:
        tuple: xlim (tuple), ylim (tuple), zlim (tuple), areas (ndarray) - areas of each face of the cuboid
    """
    size_x, size_y, size_z = active_magnet.get_size()

    xlim = (-size_x / 2, size_x / 2)
    ylim = (-size_y / 2, size_y / 2)
    zlim = (-size_z / 2, size_z / 2)

    areas = _np.array([size_y * size_z, size_x * size_z, size_x * size_y])
    return xlim, ylim, zlim, areas


def _gen_grid(xlim=(0, 10), ylim=(0, 10), num_samples=10):
    """Generates a 2D grid of N**2 points for a given pair of x and y limits.

    Args:
        xlim (tuple, optional): Min and max x values. Defaults to (0, 10).
        ylim (tuple, optional): Min and max y values. Defaults to (0, 10).
        num_samples (int, optional): Number of points to generate in each direction. Defaults to 10.

    Returns:
        ndarray: (num_samples**2, 3) array of points
    """
    xmin = xlim[0]
    xmax = xlim[1]
    ymin = ylim[0]
    ymax = ylim[1]

    dx = xmax - xmin
    dy = ymax - ymin

    xmin += dx / 2 / num_samples
    xmax -= dx / 2 / num_samples

    ymin += dy / 2 / num_samples
    ymax -= dy / 2 / num_samples

    NJ = num_samples * 1j
    x, y = _np.mgrid[xmin:xmax:NJ, ymin:ymax:NJ]
    return x.ravel(), y.ravel()


def _gen_planar_grid(
    xlim,
    ylim,
    zlim,
    center,
    face="x",
    num_samples=10,
    unit="mm",
    reverse_rotation=None,
):
    """Returns grids of points for two planes of a cuboid

    Args:
        xlim (tuple): min and max x values (float)
        ylim (tuple): min and max y values (float)
        zlim (tuple): min and max z values (float)
        face (str, optional): family of planes to generate (x, y, or z). Defaults to "x".
        num_samples (int, optional): Number of grid points to generate (10x10). Defaults to 10.
        unit (str, optional): Length scale. Defaults to "mm".

    Returns:
        tuple: points_lower, points_upper both (num_samples**2, 3) ndarrays
    """
    xc, yc, zc = center

    if face.lower() == "x":
        range_1 = ylim
        range_2 = zlim
        planes = xlim
    elif face.lower() == "y":
        range_1 = xlim
        range_2 = zlim
        planes = ylim

    elif face.lower() == "z":
        range_1 = xlim
        range_2 = ylim
        planes = zlim

    # Gets centroids of num_samples
    points_1, points_2 = _gen_grid(xlim=range_1, ylim=range_2, num_samples=num_samples)
    points_3_lower = _np.tile(planes[0], points_1.shape)
    points_3_upper = _np.tile(planes[1], points_1.shape)

    if reverse_rotation is not None:
        if face.lower() == "x":
            pos_vec = Quaternion._prepare_vector(points_3_lower, points_1, points_2)
            x, y, z = reverse_rotation * pos_vec
            points_lower = Point_Array3(x + xc, y + yc, z + zc, unit=unit)

            pos_vec = Quaternion._prepare_vector(points_3_upper, points_1, points_2)
            x, y, z = reverse_rotation * pos_vec
            points_upper = Point_Array3(x + xc, y + yc, z + zc, unit=unit)

        elif face.lower() == "y":
            pos_vec = Quaternion._prepare_vector(points_1, points_3_lower, points_2)
            x, y, z = reverse_rotation * pos_vec
            points_lower = Point_Array3(x + xc, y + yc, z + zc, unit=unit)

            pos_vec = Quaternion._prepare_vector(points_1, points_3_upper, points_2)
            x, y, z = reverse_rotation * pos_vec
            points_upper = Point_Array3(x + xc, y + yc, z + zc, unit=unit)

        elif face.lower() == "z":

            pos_vec = Quaternion._prepare_vector(points_1, points_2, points_3_lower)
            x, y, z = reverse_rotation * pos_vec
            points_lower = Point_Array3(x + xc, y + yc, z + zc, unit=unit)

            pos_vec = Quaternion._prepare_vector(points_1, points_2, points_3_upper)
            x, y, z = reverse_rotation * pos_vec
            points_upper = Point_Array3(x + xc, y + yc, z + zc, unit=unit)
    else:

        if face.lower() == "x":
            points_lower = Point_Array3(
                points_3_lower + xc, points_1 + yc, points_2 + zc, unit=unit
            )
            points_upper = Point_Array3(
                points_3_upper + xc, points_1 + yc, points_2 + zc, unit=unit
            )

        elif face.lower() == "y":
            points_lower = Point_Array3(
                points_1 + xc, points_3_lower + yc, points_2 + zc, unit=unit
            )
            points_upper = Point_Array3(
                points_1 + xc, points_3_upper + yc, points_2 + zc, unit=unit
            )

        elif face.lower() == "z":
            points_lower = Point_Array3(
                points_1 + xc, points_2 + yc, points_3_lower + zc, unit=unit
            )
            points_upper = Point_Array3(
                points_1 + xc, points_2 + yc, points_3_upper + zc, unit=unit
            )

    return points_lower, points_upper


def _calc_field_face(active_magnet, points):
    """Calculates the total force and torque acting on one face of a magnet due
    to all other instantiated magnets

    Args:
        active_magnet (Magnet3D): Target magnet
        points (Point_Array3): Grid of points (x,y,z)

    Returns:
        tuple: total_field (float), total_torque (float)
    """
    field = _allocate_field_array3(points.x, points.y, points.z)
    for magnet in Magnet3D.instances:
        if magnet is not active_magnet:
            Bx, By, Bz = magnet.get_field(points.x, points.y, points.z)
            field.x += Bx.reshape(field.x.shape)
            field.y += By.reshape(field.y.shape)
            field.z += Bz.reshape(field.z.shape)
    xc, yc, zc = active_magnet.get_center()
    field.x[~_np.isfinite(field.x)] = 0.0
    field.y[~_np.isfinite(field.y)] = 0.0
    field.z[~_np.isfinite(field.z)] = 0.0
    total_field = _np.array((_np.sum(field.x), _np.sum(field.y), _np.sum(field.z)))

    torques = _np.cross(
        _np.array(
            [points.x.ravel() - xc, points.y.ravel() - yc, points.z.ravel() - zc]
        ).T,
        _np.array([field.x.ravel(), field.y.ravel(), field.z.ravel()]).T,
    )
    total_torque = _np.sum(torques, axis=0)

    return total_field, total_torque


def calc_force_prism(active_magnet, num_samples=20, unit="mm"):
    """Calculates the total force on a cuboidal magnet due to all other instantiated magnets

    Args:
        active_magnet (Prism): Target Magnet
        num_samples (int, optional): Number of grid points per face. Defaults to 20.
        unit (str, optional): Length scale. Defaults to "mm".

    Returns:
        total_field (float), total_torque (float)
    """
    num_points_sq = num_samples * num_samples
    Jvec = active_magnet.get_Jr()
    Jnorm = active_magnet.Jr

    if _np.any(
        _np.fabs(
            _np.array(
                [
                    active_magnet.alpha_rad,
                    active_magnet.beta_rad,
                    active_magnet.gamma_rad,
                ]
            )
        )
        > active_magnet.tol
    ):

        _, reverse_rotation = active_magnet._generate_rotation_quaternions()
        Jrot = reverse_rotation * Jvec
    else:
        reverse_rotation = None
        Jrot = Jvec

    center = active_magnet.get_center()

    xlim, ylim, zlim, areas = _get_ranges_prism(active_magnet)
    force = _np.zeros(3)
    torque = _np.zeros(3)
    if _np.fabs(Jvec[0] / Jnorm) > FP_CUTOFF:
        points_lower, points_upper = _gen_planar_grid(
            xlim,
            ylim,
            zlim,
            center,
            face="x",
            num_samples=num_samples,
            unit=unit,
            reverse_rotation=reverse_rotation,
        )

        total_field, total_torque = _calc_field_face(active_magnet, points_lower)
        force += total_field * -Jrot[0] * areas[0] / num_points_sq
        torque += total_torque * -Jrot[0] * areas[0] / num_points_sq

        total_field, total_torque = _calc_field_face(active_magnet, points_upper)
        force += total_field * Jrot[0] * areas[0] / num_points_sq
        torque += total_torque * Jrot[0] * areas[0] / num_points_sq

    if _np.fabs(Jvec[1] / Jnorm) > FP_CUTOFF:
        points_lower, points_upper = _gen_planar_grid(
            xlim,
            ylim,
            zlim,
            center,
            face="y",
            num_samples=num_samples,
            unit=unit,
            reverse_rotation=reverse_rotation,
        )

        total_field, total_torque = _calc_field_face(active_magnet, points_lower)
        force += total_field * -Jrot[1] * areas[1] / num_points_sq
        torque += total_torque * -Jrot[1] * areas[1] / num_points_sq

        total_field, total_torque = _calc_field_face(active_magnet, points_upper)
        force += total_field * Jrot[1] * areas[1] / num_points_sq
        torque += total_torque * Jrot[1] * areas[1] / num_points_sq

    if _np.fabs(Jvec[2] / Jnorm) > FP_CUTOFF:
        points_lower, points_upper = _gen_planar_grid(
            xlim,
            ylim,
            zlim,
            center,
            face="z",
            num_samples=num_samples,
            unit=unit,
            reverse_rotation=reverse_rotation,
        )

        total_field, total_torque = _calc_field_face(active_magnet, points_lower)
        force += total_field * -Jrot[2] * areas[2] / num_points_sq
        torque += total_torque * -Jrot[2] * areas[2] / num_points_sq

        total_field, total_torque = _calc_field_face(active_magnet, points_upper)
        force += total_field * Jrot[2] * areas[2] / num_points_sq
        torque += total_torque * Jrot[2] * areas[2] / num_points_sq

    scaling_factor = get_unit_value_meter(points_lower.get_unit())
    force /= MU0 / scaling_factor / scaling_factor
    torque /= MU0 / scaling_factor / scaling_factor / scaling_factor

    return force, torque


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""Global Constants
PI, PI/2, PI/4, µ0, and three internally used constants:
FP_CUTOFF = 1e-8, ALIGN_CUTOFF = 1e-5, MAG_TOL = 1e-4
"""

__all__ = ["PI", "PI_2", "PI_4", "MU0", "FP_CUTOFF", "ALIGN_CUTOFF", "MAG_TOL"]

# float: Module level PI.
from math import pi as PI

# float: Module level PI/2.
PI_2 = PI / 2.0

# float: Module level PI/4.
PI_4 = PI / 4.0

# float: Module level MU0, permittivity of free space.
MU0 = 4e-7 * PI

# float: Floating point cut off used for vector norm, and logs in mesh calculations
FP_CUTOFF = 1e-8

# Alignment precision for rotation of triangles in mesh magnets
ALIGN_CUTOFF = 1e-5

# Tolerance for computation of fields due to each magnetisation component Jx, Jy, Jz
# sufficient for 0.01 degree accuracy
MAG_TOL = 1e-4


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""Plotting routines

This module contains all functions needed to plot lines and contours for 2D
magnetic sources, and 

"""

__all__ = ["plot_2D_line", "plot_2D_contour", "plot_3D_contour", "plot_sub_contour_3D"]

from .. import magnets as _mag
from ..utils._conversions import get_unit_value_meter, get_unit_value_tesla
from ..utils._vector_structs import Field2, Point_Array2

import numpy as _np
import matplotlib.pyplot as _plt
import matplotlib.cm as _cm
from matplotlib.patches import Rectangle as _Rect
from matplotlib.patches import Circle as _Circ
from matplotlib.patches import Arrow as _Arrow
from matplotlib.transforms import Affine2D


class patch(object):
    """Encodes magnet dimensions for drawing on plots"""

    def __init__(self, x, y, width, height, transform, type):
        """Initialse a patch

        Args:
            x (float): centre, x
            y (float): center, y
            width (float): width
            height (float): width
            transform (matplotlib Affine2D): transform object, `rotate_deg_around`
        """
        super().__init__()
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.transform = transform
        self.type = type

    def __repr__(self) -> str:
        return f"(x: {self.x}, y: {self.y} w:{self.width}, h: {self.height})"

    def __str__(self) -> str:
        return f"(x: {self.x}, y: {self.y} w:{self.width}, h: {self.height})"


class arrow(object):
    """Encodes magnetisation vector for drawing on plots"""

    def __init__(self, x, y, dx, dy, transform, width=3):
        """Init Arrow

        Args:
            x (float): arrow tail, x
            y (float): arrow tail, y
            dx (float): arrow head displacement, x
            dy (float): arrow head displacement, y
            transform (matplotlib Affine2D): transformation object, `translate`
            width (int, optional): Arrow width. Defaults to 3.
        """
        super().__init__()
        self.x = x
        self.y = y
        self.dx = dx
        self.dy = dy
        self.transform = transform
        self.width = width

    def __repr__(self) -> str:
        return (
            f"(x: {self.x}, y: {self.y}, dx: {self.dx}, dy: {self.dy}, w:{self.width})"
        )

    def __str__(self) -> str:
        return (
            f"(x: {self.x}, y: {self.y}, dx: {self.dx}, dy: {self.dy}, w:{self.width})"
        )


class magnet_patch(object):
    """Magnet drawing class"""

    def __init__(self, patch, arrow) -> None:
        super().__init__()
        self.patch = patch
        self.arrow = arrow

    def __str__(self) -> str:
        return self.patch.__str__() + self.arrow.__str__()


def plot_2D_line(point_array, field, **kwargs):
    """Line Plot of field from 2D magnet

    Args:
        point_array (Point_Array2): coordinates
        field (Field2): Magnetic Field

    Kwargs:
        xlab (str): xlabel
        ylab (str): ylabel
        axis_scale (str): unused
        save_fig (bool): Save to png file. Defaults to False

    Returns:
        tuple: fig, ax reference to matplotlib figure and axis objects
    """

    xlab = kwargs.pop("xlab", f"x ({point_array.unit})")
    ylab = kwargs.pop("ylab", f"B ({field.unit})")
    axis_scale = kwargs.pop("axis_scale", "equal")

    SAVE = kwargs.pop("save_fig", False)

    fig, ax = _plt.subplots(figsize=(8, 8))
    _plt.plot(point_array.x, field.n, label=r"$|\mathbf{B}|$")
    _plt.plot(point_array.x, field.x, label=r"$B_x$")
    _plt.plot(point_array.x, field.y, label=r"$B_y$")
    _plt.legend(loc="best")
    _plt.xlabel(xlab)
    _plt.ylabel(ylab)

    _plt.show()

    if SAVE:
        _plt.savefig("line_plot.png", dpi=300)
        # _plt.savefig('contour_plot.pdf', dpi=300)
    return fig, ax


def plot_2D_contour(point_array, field, **kwargs):
    """Contour plot of field

    Args:
        point_array (Point_Array2): coordinates
        field (Field2): Magnetic Field

    Kwargs:
        save_fig (bool): Save to png file. Defaults to False
        xlab (str): x axis label
        ylab (str): y axis label
        clab (str): label for colorbar
        axis_scale (str): axis aspect ratio Defaults to 'equal'.
        show_magnets (bool): Draw magnets. Defaults to True
        field_component (str): Defaults to 'n'.
        plot_type (str): Draw `contour` or `streamplot`. Defaults to 'contour'
        cmap (str): Colormap. Defaults to `viridis`
        vector_plot (bool): Draw vectors as arrows. Defaults to False.
        vector_color (str): Arrow color. Defaults to 'w'
        cmin (float): Color scale minimum. Defaults to 0.0
        cmax (float): Color scale minimum. Defaults to twice the mean field
        num_levels (int): Number of contour levels. Defaults to 11.

    Raises:
        Exception: plot_type must be 'contour' or 'streamplot'

    Returns:
        tuple: fig, ax reference to matplotlib figure and axis objects
    """
    import matplotlib.cm as _cm
    from ..magnets._polygon2D import PolyMagnet

    show_magnets = kwargs.pop("show_magnets", True)

    xlab = kwargs.pop("xlab", f"x ({point_array.unit})")
    ylab = kwargs.pop("ylab", f"y ({point_array.unit})")
    clab = kwargs.pop("clab", f"B ({field.unit})")
    axis_scale = kwargs.pop("axis_scale", "equal")

    SAVE = kwargs.pop("save_fig", False)

    field_component = kwargs.pop("field_component", "n")

    plot_type = kwargs.pop("plot_type", "contour")
    fig, ax = _plt.subplots(figsize=(8, 8))

    if plot_type.lower() == "contour":
        cmap = kwargs.pop("cmap", "viridis")
        vector_plot = kwargs.pop("vector_plot", False)
        vector_color = kwargs.pop("vector_color", "w")
        NQ = kwargs.pop("vector_arrows", 11)

        if field_component == "x":
            field_chosen = field.x
        elif field_component == "y":
            field_chosen = field.y
        else:
            field_chosen = field.n

        finite_field = field_chosen[_np.isfinite(field_chosen)]
        cmin = kwargs.pop("cmin", 0)
        cmax = kwargs.pop("cmax", round(finite_field.mean() * 2, 1))
        num_levels = kwargs.pop("num_levels", 11)

        lev2 = _np.linspace(cmin, cmax, 256, endpoint=True)

        CS = _plt.contourf(
            point_array.x,
            point_array.y,
            field_chosen,
            levels=lev2,
            cmap=_plt.get_cmap(cmap),
            extend="max",
        )

        # Draw contour lines
        if num_levels > 1:
            lev1 = _np.linspace(cmin, cmax, num_levels, endpoint=True)
            _ = _plt.contour(
                point_array.x,
                point_array.y,
                field_chosen,
                vmin=cmin,
                vmax=cmax,
                levels=lev1,
                linewidths=1.0,
                colors="k",
            )
            CB = _plt.colorbar(CS, ticks=lev1)
        else:
            CB = _plt.colorbar(CS)

        # Draw field vectors
        if vector_plot:
            _vector_plot2(point_array, field, NQ, vector_color)

    elif plot_type.lower() == "streamplot":

        xpl = point_array.x[:, 0]
        ypl = point_array.y[0, :]
        cmap = kwargs.pop("cmap", None)

        if cmap is not None:
            cmin = kwargs.pop("cmin", -round(_np.nanmean(field.n), 1))
            cmax = kwargs.pop("cmax", round(_np.nanmean(field.n), 1))
            stream_shading = kwargs.pop("stream_color", "vertical")
            norm = _cm.colors.Normalize(vmin=cmin, vmax=cmax)

            stream_dict = {
                "normal": field.n.T,
                "horizontal": field.x.T,
                "vertical": field.y.T,
            }

            CS = _plt.streamplot(
                xpl,
                ypl,
                field.x.T / field.n.T,
                field.y.T / field.n.T,
                color=stream_dict.get(stream_shading, "normal"),
                density=1.2,
                norm=norm,
                cmap=cmap,
                linewidth=0.5,
            )
            CB = _plt.colorbar(CS.lines)

        else:
            color = kwargs.pop("color", "k")
            CS = _plt.streamplot(
                xpl,
                ypl,
                field.x.T / field.n.T,
                field.y.T / field.n.T,
                density=1.2,
                linewidth=0.5,
                color=color,
            )
            CB = None

    else:
        raise Exception("plot_type must be 'contour' or 'streamplot'")

    # Draw magnets and magnetisation arrows
    if show_magnets:
        _draw_magnets2(ax)
        if len(PolyMagnet.instances) > 0:
            for magnet in PolyMagnet.instances:
                poly = _plt.Polygon(
                    _np.array(magnet.polygon.vertices),
                    ec="k",
                    fc="w",
                    zorder=5,
                )
                ax.add_patch(poly)

    if CB is not None:
        CB.ax.get_yaxis().labelpad = 15
        CB.ax.set_ylabel(clab, rotation=270)
    _plt.axis(axis_scale)
    _plt.xlabel(xlab)
    _plt.ylabel(ylab)
    _plt.show()
    fig.tight_layout()
    ax.axis("scaled")
    if SAVE:
        _plt.savefig("contour_plot.png", dpi=300)

    return fig, ax


def _num_patch_2D():
    """Generates patches and arrows for drawing

    Returns:
        tuple: (list, list) lists of patch and arrow objects
    """
    from ..magnets._magnet2D import Magnet2D, Rectangle, Circle

    patch_array = []
    for magnet in Magnet2D.instances:
        if issubclass(magnet.__class__, Rectangle):
            magnet_patch_tmp = _gen_rect_patch(magnet)
            patch_array.append(magnet_patch_tmp)
        elif issubclass(magnet.__class__, Circle):
            magnet_patch_tmp = _gen_circle_patch(magnet)
            patch_array.append(magnet_patch_tmp)

    return patch_array


def _gen_rect_patch(magnet):
    """Generates rectangular patches and arrows to represent magnets for 2D plots

    Args:
        magnet (Magnet2D): instance of a magnet class
    Returns:
        magnet_patch: magnet_patch data structure
    """
    # from matplotlib.transforms import Affine2D

    patch_tmp = patch(
        x=(magnet.center[0] - magnet.a),
        y=(magnet.center[1] - magnet.b),
        width=(2 * magnet.a),
        height=(2 * magnet.b),
        transform=Affine2D().rotate_deg_around(
            (magnet.center[0]),
            (magnet.center[1]),
            -magnet.alpha,
        ),
        type="rectangle",
    )

    Jnorm = magnet.get_Jr() / _np.abs(magnet.Jr)
    offset = _np.multiply(Jnorm, magnet.get_size()) / 2

    arrow_tmp = arrow(
        x=(magnet.center[0] - offset[0]),
        y=(magnet.center[1] - offset[1]),
        dx=(2 * offset[0]),
        dy=(2 * offset[1]),
        transform=Affine2D().translate(0, 0),
    )

    magnet_patch_tmp = magnet_patch(patch_tmp, arrow_tmp)
    return magnet_patch_tmp


def _gen_circle_patch(magnet):
    """Generates circcmaxar patches and arrows to represent magnets for 2D plots

    Args:
        magnet (Magnet2D): instance of a magnet class

    Returns:
        magnet_patch: magnet_patch data structure
    """
    # from matplotlib.transforms import Affine2D

    patch_tmp = patch(
        x=(magnet.center[0]),
        y=(magnet.center[1]),
        width=(magnet.radius),
        height=(magnet.radius),
        transform=Affine2D().rotate_deg_around(
            (magnet.center[0]),
            (magnet.center[1]),
            -magnet.alpha + magnet.phi,
        ),
        type="circle",
    )

    Jnorm = magnet.get_Jr() / _np.abs(magnet.Jr)
    offset = magnet.radius * Jnorm[0] / 2
    arrow_tmp = arrow(
        x=(magnet.center[0] - offset),
        y=(magnet.center[1]),
        dx=(2 * offset),
        dy=(2 * 0),
        transform=Affine2D().translate(0, 0),
    )
    magnet_patch_tmp = magnet_patch(patch_tmp, arrow_tmp)
    return magnet_patch_tmp


def _draw_magnets2(ax):
    """Draws Magnets and magnetisation arrows on 2D plots

    Args:
        ax (axis): axis reference
    """
    patch_array = _num_patch_2D()

    # Need original axis transform data before making additional transformations
    axis_transform = ax.transData
    for p in patch_array:
        if p.patch.type == "rectangle":
            sq = _Rect(
                xy=(p.patch.x, p.patch.y),
                width=p.patch.width,
                height=p.patch.height,
                fill=True,
                facecolor="#F5F5F5",
                edgecolor="k",
                zorder=5,
                transform=p.patch.transform + axis_transform,
            )

            ax.add_patch(sq)

        if p.patch.type == "circle":
            circ = _Circ(
                xy=(p.patch.x, p.patch.y),
                radius=p.patch.width,
                fill=True,
                facecolor="#F5F5F5",
                edgecolor="k",
                zorder=5,
                transform=p.patch.transform + axis_transform,
            )

            ax.add_patch(circ)

        ar = _Arrow(
            p.arrow.x,
            p.arrow.y,
            p.arrow.dx,
            p.arrow.dy,
            width=p.arrow.width,
            zorder=6,
            color="k",
            # must translate arrow before rotating
            transform=p.arrow.transform + p.patch.transform + axis_transform,
        )

        ax.add_patch(ar)


def _vector_plot2(points, field, NQ, vector_color):
    """Draws quiver plot of field vectors

    Args:
        points (Point_Array2): coordinates
        field (Field2): Magnetic field
        NQ (int): Plot every NQth arrow in quiver/vector plot
        vector_color (str): Color of arrows
    """
    NPx, NPy = points.x.shape
    if NQ != 0:
        NSx, NSy = NPx // NQ, NPy // NQ
        with _np.errstate(divide="ignore", invalid="ignore"):
            _plt.quiver(
                points.x[::NSx, ::NSy],
                points.y[::NSx, ::NSy],
                field.x[::NSx, ::NSy] / field.n[::NSx, ::NSy],
                field.y[::NSx, ::NSy] / field.n[::NSx, ::NSy],
                color=vector_color,
                alpha=1,
            )


def plot_3D_contour(points, field, plane, **kwargs):
    """Contour plot of field

    Args:
        points (Point_Array2): coordinates
        field (Field2): Magnetic field
        plane (str): Plane to draw contour on. Can be 'xy', 'xz', or 'yz'

    Raises:
        Exception: plot_type must be 'contour' or 'streamplot

    Returns:
        tuple: fig, ax reference to matplotlib figure and axis objects
    """
    axis_scale = kwargs.pop("axis_scale", "equal")

    plot_type = kwargs.pop("plot_type", "contour")

    xlab = kwargs.pop("xlab", "x (" + points.unit + ")")
    ylab = kwargs.pop("ylab", "y (" + points.unit + ")")
    zlab = kwargs.pop("zlab", "z (" + points.unit + ")")
    clab = kwargs.pop("clab", "B (" + field.unit + ")")

    SAVE = kwargs.pop("save_fig", False)

    finite_field = field.n[_np.isfinite(field.n)]

    cmax = kwargs.pop("cmax", round(finite_field.mean() * 2, 1))
    num_levels = kwargs.pop("num_levels", 11)

    if plane.lower() == "xy":
        plot_x = points.x
        plot_y = points.y
        plot_xlab = xlab
        plot_ylab = ylab
        stream_x = field.x
        stream_y = field.z

    elif plane.lower() == "xz":
        stream_x = field.x
        stream_y = field.z
        plot_x = points.x
        plot_y = points.z
        plot_xlab = xlab
        plot_ylab = zlab

    else:
        stream_x = field.y
        stream_y = field.z
        plot_x = points.y
        plot_y = points.z
        plot_xlab = ylab
        plot_ylab = zlab

    fig, ax = _plt.subplots(figsize=(8, 8))

    # Generate Contour Plot
    if plot_type.lower() == "contour":
        vector_plot = kwargs.pop("vector_plot", False)
        vector_color = kwargs.pop("vector_color", "w")
        NQ = kwargs.pop("vector_arrows", 11)

        cmap = kwargs.pop("cmap", "viridis")
        cmin = kwargs.pop("cmin", 0)
        lev2 = _np.linspace(cmin, cmax, 256, endpoint=True)
        CS = _plt.contourf(
            plot_x,
            plot_y,
            field.n,
            levels=lev2,
            cmap=_plt.get_cmap(cmap),
            extend="max",
        )

        # Draw contour lines
        if num_levels > 1:
            lev1 = _np.linspace(cmin, cmax, num_levels, endpoint=True)
            _ = _plt.contour(
                plot_x,
                plot_y,
                field.n,
                vmin=cmin,
                vmax=cmax,
                levels=lev1,
                linewidths=1.0,
                colors="k",
            )
            CB = _plt.colorbar(CS, ticks=lev1)

        else:
            CB = _plt.colorbar(CS)

        if vector_plot:
            B_2D = Field2(stream_x, stream_y, unit=field.unit)
            B_2D.n = field.n
            points_2D = Point_Array2(plot_x, plot_y, unit=points.unit)
            _vector_plot2(points_2D, B_2D, NQ, vector_color)

    # Generates streamplot
    elif plot_type.lower() == "streamplot":
        xpl = plot_x[:, 0]
        ypl = plot_y[0, :]
        cmap = kwargs.pop("cmap", None)
        if cmap is not None:
            cmin = kwargs.pop("cmin", -round(finite_field.mean() * 2, 1))
            cmax = kwargs.pop("cmax", round(finite_field.mean() * 2, 1))

            stream_shading = kwargs.pop("stream_shading", "vertical")
            norm = _cm.colors.Normalize(vmin=cmin, vmax=cmax)

            stream_dict = {
                "normal": field.n.T,
                "horizontal": stream_x.T,
                "vertical": stream_y.T,
            }

            CS = _plt.streamplot(
                xpl,
                ypl,
                stream_x.T / field.n.T,
                stream_y.T / field.n.T,
                color=stream_dict.get(stream_shading, "normal"),
                density=1.2,
                norm=norm,
                cmap=cmap,
                linewidth=0.5,
            )
            CB = _plt.colorbar(CS.lines)
        else:
            color = kwargs.pop("color", "k")
            CS = _plt.streamplot(
                xpl,
                ypl,
                stream_x.T / field.n.T,
                stream_y.T / field.n.T,
                density=1.2,
                linewidth=0.5,
                color=color,
            )
            CB = None

    else:
        raise Exception("plot_type must be 'contour' or 'streamplot'")

    if CB is not None:
        CB.ax.get_yaxis().labelpad = 15
        CB.ax.set_ylabel(clab, rotation=270)
    _plt.xlabel(plot_xlab)
    _plt.ylabel(plot_ylab)
    _plt.axis(axis_scale)

    if SAVE:
        _plt.savefig("contour_plot.png", dpi=300)

    return fig, ax


def plot_sub_contour_3D(plot_x, plot_y, plot_B, **kwargs):
    """Contour plot of a single magnetic field component of a 3D simulation

    Args:
        plot_x (ndarray): coordinates for x-axis of plot
        plot_y (ndarray): coordinates for y-axis of plot
        plot_B (ndarray): Magnetic field component to plot

    Returns:
        tuple: fig, ax reference to matplotlib figure and axis objects
    """
    cmap = kwargs.pop("cmap", "seismic")
    xlab = kwargs.pop("xlab", f"x (m)")
    ylab = kwargs.pop("ylab", f"y (m)")
    clab = kwargs.pop("clab", f"B (T)")

    axis_scale = kwargs.pop("axis_scale", "equal")

    SAVE = kwargs.pop("save_fig", False)

    cmin = kwargs.pop("cmin", -0.5)
    cmax = kwargs.pop("cmax", 0.5)
    num_levels = kwargs.pop("num_levels", 11)

    lev2 = _np.linspace(cmin, cmax, 256, endpoint=True)
    fig, ax = _plt.subplots(figsize=(8, 8))
    CS = _plt.contourf(
        plot_x, plot_y, plot_B, levels=lev2, cmap=_plt.get_cmap(cmap), extend="both"
    )

    if num_levels > 1:
        lev1 = _np.linspace(cmin, cmax, num_levels, endpoint=True)
        _ = _plt.contour(
            plot_x,
            plot_y,
            plot_B,
            vmin=cmin,
            vmax=cmax,
            levels=lev1,
            linewidths=1.0,
            colors="k",
        )
        CB = _plt.colorbar(CS, ticks=lev1)
    else:
        CB = _plt.colorbar(CS)

    CB.ax.get_yaxis().labelpad = 15
    CB.ax.set_ylabel(clab, rotation=270)
    _plt.xlabel(xlab)
    _plt.ylabel(ylab)
    _plt.axis("equal")
    _plt.show()

    return fig, ax


def line_plot_cylinder(magnet, **kwargs):
    """Calculates and plots the magnetic field along the central axis
    of a cylinder

    This is an example helper function.

    Args:
        magnet (Cylinder): instance of magnetic cylinder

    Returns:
        tuple: fig, ax reference to matplotlib figure and axis objects
    """
    rho = _np.linspace(-2 * magnet.radius, 2 * magnet.radius, 51)
    z = _np.array([magnet.length * 1.1 / 2])

    Br, Bz = magnet._calcB_cyl(rho, z)
    fig, ax = _plt.subplots(figsize=(8, 8))
    _plt.plot(rho * 1, Bz, label=r"$B_z$")
    _plt.plot(rho * 1, Br, label=r"$B_r$")
    _plt.legend(loc="best")
    _plt.show()
    return fig, ax


def contour_plot_cylinder(magnet, **kwargs):
    """Calculates and plots the magnetic field
    of a cylinder

    This is an example helper function.


    Args:
        magnet (Cylinder): instance of magnetic cylinder

    Returns:
        tuple: fig, ax reference to matplotlib figure and axis objects
    """
    NP = 101
    NPJ = NP * 1j
    rho, z = _np.mgrid[
        -3 * magnet.radius : 3 * magnet.radius : NPJ,
        -magnet.length : magnet.length : NPJ,
    ]
    Br, Bz = magnet._calcB_cyl(rho, z)
    Bn = _np.sqrt(Bz ** 2 + Br ** 2)

    xlab = f"r (m)"
    ylab = "z (m)"

    # plot_B = Bn
    clab = r"$|B|$ (T)"
    cmap = "viridis"
    fig, ax = plot_sub_contour_3D(
        rho * 1,
        z * 1,
        Bn,
        xlab=xlab,
        ylab=ylab,
        clab=clab,
        cmap=cmap,
        cmin=0,
        cmax=1.0,
    )
    return fig, ax



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""Routines for Three Dimensional Magnet Classes
"""
__all__ = ["get_field_3D", "grid3D", "slice3D"]

import numpy as _np
from ._vector_structs import Field3, Point_Array3


def grid3D(xmax, ymax, zmax, **kwargs):
    """Generates grid of x and y points

    Args:
        xmax (float): maximum x value
        ymax (float): maximum y value
        zmax (float): maximum y value

    Kwargs:
        num_points (int): Number of points in each direction. Defaults to 100
        xmin (float): minimum x value. Defaults to -xmax
        ymin (float): minimum y value. Defaults to -ymax
        zmin (float): minimum y value. Defaults to -zmax
        unit (string): unit length. Defaults to 'mm'

    Returns:
        Point_Array2: array of x and y values of shape (num_points, num_points) and associated unit
    """
    num_points = kwargs.pop("num_points", None)

    xmin = kwargs.pop("xmin", -1 * xmax)
    ymin = kwargs.pop("ymin", -1 * ymax)
    zmin = kwargs.pop("zmin", -1 * zmax)
    unit = kwargs.pop("unit", "mm")
    NPJ = num_points * 1j

    if num_points is None:
        num_points_x = kwargs.pop("num_points_x", 100)
        num_points_y = kwargs.pop("num_points_y", 100)
        num_points_z = kwargs.pop("num_points_z", 100)
    else:
        num_points_x = num_points
        num_points_y = num_points
        num_points_z = num_points

    x, y, z = _np.mgrid[
        xmin : xmax : num_points_x * 1j,
        ymin : ymax : num_points_y * 1j,
        zmin : zmax : num_points_z * 1j,
    ]

    return Point_Array3(x, y, z, unit=unit)


def slice3D(plane="xy", max1=1.0, max2=1.0, slice_value=0.0, unit="mm", **kwargs):
    """Generates a planar slice of values

    Args:
        plane (str, optional): plane. Defaults to "xy".
        max1 (float, optional): maximum along axis 1. Defaults to 1.0.
        max2 (float, optional): maximum along axis 2. Defaults to 1.0.
        slice_value (float, optional): constant value for third axis. Defaults to 0.0.
        unit (str, optional): length scale units. Defaults to "mm".

    Kwargs:
        num_points (int): Number of points in each direction. Defaults to 100
        min1 (float): minimum along axis 1. Defaults to -min1
        min2 (float): minimum along axis 2. Defaults to -min2

    Raises:
        Exception: plane type, must be one of 'xy', 'xz, 'yz', or 'custom'

    Returns:
        Point_Array3: array of x, y, and z values of shape (num_points, num_points) and associated unit
    """
    num_points = kwargs.pop("num_points", 100)
    min1 = kwargs.pop("min1", -1 * max1)
    min2 = kwargs.pop("min2", -1 * max2)
    NPj = num_points * 1j

    if plane.lower() == "xy":
        x, y = _np.mgrid[min1:max1:NPj, min2:max2:NPj]
        z = _np.asarray([slice_value])
        z = _np.tile(z, x.shape)

    elif plane.lower() == "xz":
        x, z = _np.mgrid[min1:max1:NPj, min2:max2:NPj]
        y = _np.asarray([slice_value])
        y = _np.tile(y, x.shape)

    elif plane.lower() == "yz":
        y, z = _np.mgrid[min1:max1:NPj, min2:max2:NPj]
        x = _np.asarray([slice_value])
        x = _np.tile(x, y.shape)

    elif plane.lower() == "custom":
        x = kwargs.pop("custom_x", _np.array([0.0]))
        y = kwargs.pop("custom_y", _np.array([0.0]))
        z = kwargs.pop("custom_z", _np.array([0.0]))

    else:
        raise Exception("plane must be one of 'xy', 'xz, 'yz', or 'custom'")

    return Point_Array3(x, y, z, unit=unit)


def get_field_3D(points):
    """Calculates magnetic field at an array of points due to every instantated
    `Magnet3D` magnet.

    Args:
        Point_Array3 (Point_Array3): array of x,y,z points and associated unit, defaults to 'mm'

    Returns:
        Field3: array of Bx,By,Bz,|B| values and associated unit (defaults to 'T')
    """
    from ..magnets import Magnet3D

    B = _allocate_field_array3(points.x, points.y, points.z)

    for magnet in Magnet3D.instances:
        Bx, By, Bz = magnet.get_field(points.x, points.y, points.z)
        B.x += Bx.reshape(B.x.shape)
        B.y += By.reshape(B.y.shape)
        B.z += Bz.reshape(B.z.shape)

    B.calc_norm()
    return B


def _allocate_field_array3(x, y, z):
    """Allocates empty Field3 data structure

    Args:
        x (ndarray): x co-ordinates
        y (ndarray): y co-ordinates
        z (ndarray): z co-ordinates

    Returns:
        Field3: Empty data structure
    """

    # Ensure x,y,z are numpy arrays (even of element 1)
    x = _np.atleast_1d(x)
    y = _np.atleast_1d(y)
    z = _np.atleast_1d(z)

    # Determine array shape:
    if _np.ndim(x) == 3:  # Volume meshgrid
        B = Field3(_np.zeros_like(x), _np.zeros_like(x), _np.zeros_like(x))

    elif _np.ndim(x) == 2:  # planar slice
        B = Field3(_np.zeros_like(x), _np.zeros_like(x), _np.zeros_like(x))

    elif _np.ndim(y) == 2:  # planar slice
        B = Field3(_np.zeros_like(y), _np.zeros_like(y), _np.zeros_like(y))

    else:  # line or single point
        B = Field3(
            _np.zeros(max(x.size, y.size, z.size)),
            _np.zeros(max(x.size, y.size, z.size)),
            _np.zeros(max(x.size, y.size, z.size)),
        )
    return B


def _get_max_array(list_arrays):
    """Gets the shape and size of the largest array in a list

    Args:
        list_arrays (list): list of numpy ndarrays

    Returns:
        tuple: max_shape (tuple), max_size (int)
    """
    max_shape = (1, 1)
    max_size = 0
    for array in list_arrays:
        if array.size > max_size:
            max_shape = array.shape
            max_size = array.size
    return max_shape, max_size


def _check_tile_array(array, max_size, max_shape):
    if array.size < max_size:
        return _np.tile(array, max_shape)
    else:
        return array


def _tile_arrays(x, y, z):
    """Tiles arrays to match the largest array of x, y, z

    Args:
        x (ndarray/float): x-coordinates
        y (ndarray/float): y-coordinates
        z (ndarray/float): z-coordinates

    Returns:
        tuple: x, y, z ndarrays
    """
    x = _np.asarray(x)
    y = _np.asarray(y)
    z = _np.asarray(z)

    max_shape, max_size = _get_max_array([x, y, z])

    x = _check_tile_array(x, max_size, max_shape)
    y = _check_tile_array(y, max_size, max_shape)
    z = _check_tile_array(z, max_size, max_shape)

    return x, y, z


def _apply_mask(magnet, field, mask):
    """Calculates B, or applies Nan, inside magnets

    Args:
        magnet (Magnet3D): instantiated magnet
        field (Field3): magnetic field vector
        mask (array): boolean array where True marks coordinates inside the magnet

    Returns:
        Field3: masked magnetic field vector
    """
    from ..magnets import Prism, Cylinder, Sphere

    J = magnet.get_Jr()
    mask_magnet = magnet._mask_magnet

    if mask_magnet:
        field.x[mask] = _np.NaN
        field.y[mask] = _np.NaN
        field.z[mask] = _np.NaN

    else:
        if issubclass(magnet.__class__, Prism):
            field.x[mask] += J[0]
            field.y[mask] += J[1]
            field.z[mask] += J[2]

        elif issubclass(magnet.__class__, Cylinder):
            pass

        elif issubclass(magnet.__class__, Sphere):
            field.x[mask] = 0.0
            field.y[mask] = 0.0
            field.z[mask] = magnet.Jr * 2 / 3

    return field


# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""pymagnets.utils

This module imports the classes and functions in the private modules to create a public API, including:

- Quaternion()
- Global Constants
- Point structures
- Vector structures

"""

from ._quaternion import *
from .global_const import *
from ._point_structs import *
from ._routines2D import *
from ._routines3D import *
from ._trigonometry3D import *

# Not implemented yet:
# from ._assemblies2D import *
# from ._fit import *


__all__ = ["calc_force_sphere"]

import numpy as _np
from ..utils.global_const import FP_CUTOFF, MU0, PI
from ..utils._conversions import get_unit_value_meter
from ..utils._vector_structs import Point_Array3
from ..utils._routines3D import _allocate_field_array3
from ..utils._quaternion import Quaternion
from ..magnets import Magnet3D


def _gen_sphere_grid(active_magnet, Jvec, num_samples=10, unit="mm"):

    NS = num_samples * 1j
    radius = active_magnet.get_size()
    # Jr = active_magnet.Jr

    u, v = _np.mgrid[0 : 2 * PI * (num_samples - 1) / num_samples : NS, 0:PI:NS]

    x = radius * _np.cos(u) * _np.sin(v)
    y = radius * _np.sin(u) * _np.sin(v)
    z = radius * _np.cos(v)
    # Jnorm = Jr * _np.cos(v)

    # Delete origin duplicates
    # Jnorm = _np.delete(Jnorm, _np.arange(num_samples, Jnorm.size, num_samples)).ravel()

    y = _np.delete(y, _np.arange(num_samples, x.size, num_samples))
    z = _np.delete(z, _np.arange(num_samples, x.size, num_samples))
    x = _np.delete(x, _np.arange(num_samples, x.size, num_samples))
    Jnorm = _np.dot(Jvec, _np.array([x.ravel(), y.ravel(), z.ravel()]) / radius)

    points = Point_Array3(x.ravel(), y.ravel(), z.ravel(), unit=unit)

    return points, Jnorm


def _calc_field_face_sphere(active_magnet, points):
    field = _allocate_field_array3(points.x, points.y, points.z)
    for magnet in Magnet3D.instances:
        if magnet is not active_magnet:
            Bx, By, Bz = magnet.get_field(points.x, points.y, points.z)
            field.x += Bx
            field.y += By
            field.z += Bz
    xc, yc, zc = active_magnet.get_center()
    field.x[~_np.isfinite(field.x)] = 0.0
    field.y[~_np.isfinite(field.y)] = 0.0
    field.z[~_np.isfinite(field.z)] = 0.0
    fields = _np.array((field.x, field.y, field.z))

    torques = _np.cross(
        _np.array([points.x - xc, points.y - yc, points.z - zc]).T,
        _np.array([field.x, field.y, field.z]).T,
    ).T
    return fields, torques


def calc_force_sphere(active_magnet, num_samples=200, unit="mm"):

    Jvec = active_magnet.get_Jr()
    Jnorm = active_magnet.Jr

    if _np.any(
        _np.fabs(
            _np.array(
                [
                    active_magnet.alpha_rad,
                    active_magnet.beta_rad,
                    active_magnet.gamma_rad,
                ]
            )
        )
        > active_magnet.tol
    ):

        _, reverse_rotation = active_magnet._generate_rotation_quaternions()
        Jrot = reverse_rotation * Jvec
    else:
        reverse_rotation = None
        Jrot = Jvec

    xc, yc, zc = active_magnet.get_center()

    points, Jn = _gen_sphere_grid(
        active_magnet,
        Jvec=Jrot,
        num_samples=num_samples,
        unit=unit,
    )
    points.x += xc
    points.y += yc
    points.z += zc

    area = 4 * PI * active_magnet.radius ** 2

    fields, torques = _calc_field_face_sphere(active_magnet, points)
    forces = Jn * fields

    torques = Jn * torques

    num_samples = _np.size(points.x)

    force = _np.sum(forces, axis=1) * area / num_samples
    torque = _np.sum(torques, axis=1) * area / num_samples

    scaling_factor = get_unit_value_meter(points.get_unit())
    force /= MU0 / scaling_factor / scaling_factor
    torque /= MU0 / scaling_factor / scaling_factor / scaling_factor
    return force, torque


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""2D Magnet Classes

This private module implements the `Rectangle` and `Square` and `Circle` 2D magnet classes. The parent class `Magnet2D` implements the location and orientation
methods, i.e. magnet center and quaternion methods for rotating the magnet with respect
to each principal axis.

"""
import numpy as _np
from ._magnet_base import Magnet
from ..utils.global_const import MAG_TOL, PI

__all__ = ["Magnet2D", "Rectangle", "Square", "Circle"]


class Magnet2D(Magnet):
    """2D Magnet Base Class"""

    mag_type = "Magnet2D"

    def __init__(self, Jr, **kwargs) -> None:
        """Init Method

        Args:
            Jr (float): signed magnetised of remnant magnetisationnega

        Kwargs:
            alpha (float): Magnetisation orientation angle (in degrees). Defaults to 0.
            center (tuple or ndarrray): magnet center (x, y). Defaults to (0,0).
        """
        super().__init__()
        self.Jr = Jr

        # Magnet rotation w.r.t. x-axis
        self.alpha = kwargs.pop("alpha", 0.0)
        self.alpha_radians = _np.deg2rad(self.alpha)

        self.center = kwargs.pop("center", _np.array([0.0, 0.0]))
        self.center = _np.asarray(self.center)

    def get_center(self):
        """Returns magnet centre

        Returns:
            center (ndarray): numpy array [xc, yc]
        """
        return self.center

    def get_orientation(self):
        """Returns magnet orientation, `alpha` in degrees

        Returns:
            float: alpha, rotation angle w.r.t x-axis.
        """

        return self.alpha

    def get_field(self):
        """Calculates the magnetic field.

        This is a template that needs to be implemented for each magnet
        """
        pass

    def get_force_torque(self):
        """Calculates the force and torque on a magnet due to all other magnets.

        This is a template that needs to be implemented for each magnet.
        """
        pass


class Rectangle(Magnet2D):
    """Rectangular 2D Magnet Class"""

    mag_type = "Rectangle"

    def __init__(self, width=20.0, height=40.0, Jr=1.0, **kwargs):
        """Init Method

        Args:
            width (float, optional): Magnet Width. Defaults to 20.0.
            height (float, optional): Magnet Height. Defaults to 40.0.
            Jr (float, optional): Remnant Magnetisation. Defaults to 1.0.

        Kwargs:
            alpha (float): Magnetisation orientation angle (in degrees). Defaults to 0.
            center (tuple or ndarrray): magnet center (x, y). Defaults to (0,0).
            phi (float): Rotation Angle (in degrees) of magnet w.r.t x-axis. Defaults to 90.
        """
        super().__init__(Jr, **kwargs)
        self.width = width
        self.height = height

        self.a = width / 2
        self.b = height / 2

        self.phi = kwargs.pop("phi", 90)
        self.phi_rad = _np.deg2rad(self.phi)

        self.Jx = _np.around(Jr * _np.cos(self.phi_rad), decimals=6)
        self.Jy = _np.around(Jr * _np.sin(self.phi_rad), decimals=6)
        self.tol = MAG_TOL  # sufficient for 0.01 degree accuracy

    def get_size(self):
        """Returns magnet dimesions

        Returns:
            ndarray: numpy array [width, height]
        """
        return _np.array([self.width, self.height])

    def __str__(self):
        str = (
            f"{self.__class__.mag_type}\n"
            + f"J: {self.get_Jr()} (T)\n"
            + f"Size: {self.get_size()}\n"
            + f"Center {self.get_center()}\n"
            + f"Orientation: alpha {self.get_orientation()}\n"
        )
        return str

    def __repr__(self):
        str = (
            f"{self.__class__.mag_type}\n"
            + f"J: {self.get_Jr()} (T)\n"
            + f"Size: {self.get_size()}\n"
            + f"Center {self.get_center()}\n"
            + f"Orientation: alpha {self.get_orientation()}\n"
        )
        return str

    def get_Jr(self):
        """Returns Magnetisation vector

        Returns:
            ndarray: [Jx, Jy]
        """
        return _np.array([self.Jx, self.Jy])

    def get_field(self, x, y):
        """Calculates the magnetic field at point(s) x,y due to a rectangular magnet

        Args:
            x (ndarray): x co-ordinates
            y (ndarray): y co-ordinates

        Returns:
            tuple: magnetic field vector Bx (ndarray), By (ndarray)
        """
        from ..utils._routines2D import rotate_points_2D, _get_field_array_shape2

        array_shape = _get_field_array_shape2(x, y)
        Bx, By = _np.zeros(array_shape), _np.zeros(array_shape)

        if _np.fabs(self.alpha_radians) > Magnet2D.tol:
            xi, yi = rotate_points_2D(
                x - self.center[0], y - self.center[1], self.alpha_radians
            )

        # Calculate field due to x-component of magnetisation
        if _np.fabs(self.Jx / self.Jr) > Magnet2D.tol:
            if _np.fabs(self.alpha_radians) > Magnet2D.tol:

                # Calculate fields in local frame
                Btx = self._calcBx_mag_x(xi, yi)
                Bty = self._calcBy_mag_x(xi, yi)

                # Rotate fields to global frame
                Bx, By = rotate_points_2D(Btx, Bty, 2 * PI - self.alpha_radians)

            else:
                Bx = self._calcBx_mag_x(x - self.center[0], y - self.center[1])
                By = self._calcBy_mag_x(x - self.center[0], y - self.center[1])

        # Calculate field due to y-component of magnetisation
        if _np.fabs(self.Jy / self.Jr) > Magnet2D.tol:
            if _np.fabs(self.alpha_radians) > Magnet2D.tol:

                Btx = self._calcBx_mag_y(xi, yi)
                Bty = self._calcBy_mag_y(xi, yi)

                Bxt, Byt = rotate_points_2D(Btx, Bty, 2 * PI - self.alpha_radians)
                Bx += Bxt
                By += Byt
            else:

                Bx += self._calcBx_mag_y(x - self.center[0], y - self.center[1])
                By += self._calcBy_mag_y(x - self.center[0], y - self.center[1])
        return Bx, By

    def _calcBx_mag_x(self, x, y):
        """Bx using 2D Model for rectangular sheets magnetised in x-plane

        Args:
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates

        Returns:
            ndarray: Bx, x component of magnetic field
        """
        a = self.a
        b = self.b
        J = self.Jx
        # Hide the warning for situtations where there is a divide by zero.
        # This returns a NaN in the array, which is ignored for plotting.
        with _np.errstate(divide="ignore", invalid="ignore"):
            return (J / (2 * PI)) * (
                _np.arctan2((2 * a * (b + y)), (x ** 2 - a ** 2 + (y + b) ** 2))
                + _np.arctan2((2 * a * (b - y)), (x ** 2 - a ** 2 + (y - b) ** 2))
            )

    def _calcBy_mag_x(self, x, y):
        """By using 2D Model for rectangular sheets magnetised in x-plane

        Args:
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates

        Returns:
            ndarray: By, x component of magnetic field
        """
        a = self.a
        b = self.b
        J = self.Jx
        # Hide the warning for situtations where there is a divide by zero.
        # This returns a NaN in the array, which is ignored for plotting.
        with _np.errstate(divide="ignore", invalid="ignore"):
            return (-J / (4 * PI)) * (
                _np.log(((x - a) ** 2 + (y - b) ** 2) / ((x + a) ** 2 + (y - b) ** 2))
                - _np.log(((x - a) ** 2 + (y + b) ** 2) / ((x + a) ** 2 + (y + b) ** 2))
            )

    def _calcBx_mag_y(self, x, y):
        """Bx using 2D Model for rectangular sheets magnetised in y-plane

        Args:
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates

        Returns:
            ndarray: Bx, x component of magnetic field
        """
        a = self.a
        b = self.b
        J = self.Jy
        # Hide the warning for situtations where there is a divide by zero.
        # This returns a NaN in the array, which is ignored for plotting.
        with _np.errstate(divide="ignore", invalid="ignore"):
            return (J / (4 * PI)) * (
                _np.log(((x + a) ** 2 + (y - b) ** 2) / ((x + a) ** 2 + (y + b) ** 2))
                - _np.log(((x - a) ** 2 + (y - b) ** 2) / ((x - a) ** 2 + (y + b) ** 2))
            )

    def _calcBy_mag_y(self, x: float, y: float) -> float:
        """Bx using 2D Model for rectangular sheets magnetised in y-plane

        Args:
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates

        Returns:
            ndarray: By, x component of magnetic field
        """
        a = self.a
        b = self.b
        J = self.Jy
        return (J / (2 * PI)) * (
            _np.arctan2((2 * b * (x + a)), ((x + a) ** 2 + y ** 2 - b ** 2))
            - _np.arctan2((2 * b * (x - a)), ((x - a) ** 2 + y ** 2 - b ** 2))
        )


class Square(Rectangle):
    """Square 2D Magnet Class"""

    mag_type = "Square"

    def __init__(self, width=20, Jr=1.0, **kwargs):
        """Init Method

        Args:
            width (float, optional): Sqaure side length. Defaults to 20.0.
            Jr (float, optional): Remnant Magnetisation. Defaults to 1.0.

        Kwargs:
             alpha (float): Magnetisation orientation angle (in degrees). Defaults to 0.
            center (tuple or ndarrray): magnet center (x, y). Defaults to (0,0).
            phi (float): Rotation Angle (in degrees) of magnet w.r.t x-axis. Defaults to 90.
        """
        super().__init__(width=width, height=width, Jr=Jr, **kwargs)


class Circle(Magnet2D):
    """Circle 2D Magnet Class"""

    mag_type = "Circle"

    def __init__(
        self,
        radius=10,
        Jr=1.0,  # local magnetisation
        **kwargs,
    ):
        """Init Method

        Args:
            radius (float, optional): Radius. Defaults to 10.0.
            Jr (float, optional): Remnant magnetisation. Defaults to 1.0.

        Kwargs:
            alpha (float): Unused. For rotations use phi instead
            center (tuple or ndarrray): magnet center (x, y). Defaults to (0,0)
            phi (float): Rotation Angle (in degrees) of magnet w.r.t x-axis. Defaults to 90.
        """
        super().__init__(Jr, **kwargs)
        self.radius = radius
        self.phi = kwargs.pop("phi", 0)
        self.phi_rad = _np.deg2rad(self.phi)

        self.Jx = _np.around(Jr * _np.cos(self.phi_rad), decimals=6)
        self.Jy = _np.around(Jr * _np.sin(self.phi_rad), decimals=6)
        self.tol = MAG_TOL  # sufficient for 0.01 degree accuracy

        self.center = kwargs.pop("center", _np.array([0.0, 0.0]))
        self.center = _np.asarray(self.center)

    def __str__(self):
        str = (
            f"{self.__class__.mag_type}\n"
            + f"J: {self.get_Jr()} (T)\n"
            + f"Size: {self.get_size()}\n"
            + f"Center {self.get_center()}\n"
            + f"Orientation: alpha {self.get_orientation()}\n"
        )
        return str

    def __repr__(self):
        str = (
            f"{self.__class__.mag_type}\n"
            + f"J: {self.get_Jr()} (T)\n"
            + f"Size: {self.get_size()}\n"
            + f"Center {self.get_center()}\n"
            + f"Orientation: alpha {self.get_orientation()}\n"
        )
        return str

    def get_size(self):
        """Returns radius

        Returns:
            ndarray: radius
        """
        return _np.array([self.radius])

    def get_Jr(self):
        """Returns signed remnant magnetisation

        Returns:
            ndarray: remnant magnetisation
        """
        return _np.array([self.Jx, self.Jy])

    def get_field(self, x, y):
        """Calculates the magnetic field due to long bipolar cylinder

        Args:
            x (ndarray): x coordinates
            y (narray): y coordinates

        Returns:
            tuple: Bx, By magnetic field in cartesian coordinates
        """
        from ..utils._conversions import cart2pol, vector_pol2cart
        from ..utils._routines2D import rotate_points_2D

        if _np.fabs(self.alpha_radians) > Magnet2D.tol:
            xi, yi = rotate_points_2D(
                x - self.center[0], y - self.center[1], self.alpha_radians
            )

            rho, phi = cart2pol(xi, yi)

            Brho, Bphi = self._calcB_polar(rho, phi - self.phi_rad)

            # Convert magnetic fields from cylindrical to cartesian
            Bx, By = vector_pol2cart(Brho, Bphi, phi)
            Bx, By = rotate_points_2D(Bx, By, 2 * PI - self.alpha_radians)
            return Bx, By

        rho, phi = cart2pol(x - self.center[0], y - self.center[1])

        Brho, Bphi = self._calcB_polar(rho, phi - self.phi_rad)

        # Convert magnetic fields from cylindrical to cartesian
        Bx, By = vector_pol2cart(Brho, Bphi, phi)

        return Bx, By

    def _calcB_polar(self, rho, phi):
        """Calculates the magnetic field due to long bipolar cylinder in polar
        coordinates

        Args:
            rho (ndarray): radial values
            phi (ndarray): azimuthal values

        Returns:
            tuple: Br, Bphi magnetic field in polar coordinates
        """
        prefac = self.Jr * (self.radius ** 2 / rho ** 2) / 2

        Brho = prefac * _np.cos(phi)
        Bphi = prefac * _np.sin(phi)

        return Brho, Bphi



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""Example plots
"""


# def main():
#     print("Line Plot through centre of a cylinder")
#
#     import pymagnet as pm
#
#     R = 5
#     L = 25
#     m_cyl = pm.magnets.Cylinder(radius=R, length=L, Jr=1.2, center=(0.0, 0.0, 0))
#     pm.plot.plot_1D_field(m_cyl)
#
#     print("Contour Plot of 2D Magnets")
#     pm.reset_magnets()
#     cmap = "viridis"  # colormap
#     width = 20
#     height = 20
#     hgap_x = width / 2
#
#     center = (-width / 2 - hgap_x, 0)
#     _ = pm.magnets.Rectangle(
#         width=width, height=height, Jr=1.0, center=center, theta=0.0
#     )
#     center = (width / 2 + hgap_x, 0)
#     _ = pm.magnets.Rectangle(
#         width=width, height=height, Jr=1.0, center=center, theta=90.0
#     )
#
#     x, y = pm.grid2D(2 * width, 2 * height)
#     B = pm.get_field_2D(x, y)
#
#     pm.plot.plot_2D_contour(x, y, B, UL=0.6, NL=7, vector_plot=True, cmap=cmap)
#
#
# if __name__ == "__main__":
#     main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v.2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""pymagnets.utils._vector_structs

Private module consiting of vector and point array classes and their methods.
"""

import numpy as _np
from ._conversions import get_unit_value_meter, get_unit_value_tesla

__all__ = ["Field1", "Field2", "Field3", "Point_Array2", "Point_Array3"]


class Point_Array1(object):
    """1D point structure
    This is used to contain one position array (z), and the units
    ('mm', 'cm', etc)
    """

    def __init__(self, z, unit="mm"):
        """Init method

        Args:
            z (ndarray): z coordinates
            unit (str, optional): Unit of length. Defaults to "mm".

        Raises:
            ValueError: Unit must an SI prefix, e.g. km, m, cm, mm
        """
        self.z = _np.asarray(z)
        if get_unit_value_meter(unit) is not None:
            self.unit = unit
        else:
            raise ValueError("Error, not an SI prefix, e.g. km, m, cm, mm ")

    def __repr__(self) -> str:
        return f"[(Unit:{self.unit}) Array:{self.z}]"

    def __str__(self) -> str:
        return f"[(Unit:{self.unit}) Array:{self.z}]"

    def get_unit(self):
        """Gets unit

        Returns:
            str: unit
        """
        return self.unit

    def change_unit(self, new_unit, get_unit_value=get_unit_value_meter):
        """Converts point array to a different unit. e.g from 'cm' to 'mm'

        Args:
            new_unit (str): unit to be converted to
            get_unit_value (function, optional): Function for checking unit type. Defaults to get_unit_value_meter.
        """
        from ..magnets import Magnet, Prism, Cylinder

        current_unit_val = get_unit_value(self.get_unit())
        new_unit_val = get_unit_value(new_unit)
        scale_val = current_unit_val / new_unit_val

        self.z *= scale_val

        self.unit = new_unit
        for magnet in Magnet.instances:
            magnet.center = scale_val * magnet.center
            if issubclass(magnet.__class__, Prism):
                magnet.a = magnet.a * scale_val
                magnet.b = magnet.b * scale_val
                magnet.c = magnet.c * scale_val
                magnet.width = magnet.width * scale_val
                magnet.depth = magnet.depth * scale_val
                magnet.height = magnet.height * scale_val
            elif issubclass(magnet.__class__, Cylinder):
                magnet.radius = magnet.radius * scale_val
                magnet.length = magnet.length * scale_val


class Point_Array2(object):
    """2D point structure
    This is used to contain two position arrays (x, y), and the units
    ('mm', 'cm', etc)
    """

    def __init__(self, x, y, unit="mm"):
        """Init Method

        Args:
            x (ndarray): x coordinates
            y (ndarray): y coordinates
            unit (str, optional): Unit of length. Defaults to "mm".

        Raises:
            ValueError: Unit must an SI prefix, e.g. km, m, cm, mm
        """
        self.x = _np.asarray(x)
        self.y = _np.asarray(y)
        if get_unit_value_meter(unit) is not None:
            self.unit = unit
        else:
            raise ValueError("Error, not an SI prefix, e.g. km, m, cm, mm ")

    def __repr__(self) -> str:
        return f"[Unit: {self.unit}\nx: {self.x}\ny: {self.y}]"

    def __str__(self) -> str:
        return f"[Unit: {self.unit}\nx: {self.x}\ny: {self.y}]"

    def get_unit(self):
        return self.unit

    def change_unit(self, new_unit, get_unit_value=get_unit_value_meter):
        """Converts point array to a different unit. e.g from 'cm' to 'mm'

        Args:
            new_unit (str): unit to be converted to
            get_unit_value (function, optional): Function for checking unit type. Defaults to get_unit_value_meter.
        """
        from ..magnets import Magnet, Rectangle, Square, Circle, PolyMagnet

        current_unit_val = get_unit_value(self.get_unit())
        new_unit_val = get_unit_value(new_unit)
        scale_val = current_unit_val / new_unit_val

        self.x *= scale_val
        self.y *= scale_val
        self.unit = new_unit
        for magnet in Magnet.instances:
            magnet.center = scale_val * magnet.center
            if issubclass(magnet.__class__, Rectangle):
                magnet.a = magnet.a * scale_val
                magnet.b = magnet.b * scale_val
                magnet.width = magnet.width * scale_val
                magnet.height = magnet.height * scale_val
            elif issubclass(magnet.__class__, Square):
                magnet.a = magnet.a * scale_val
                magnet.width = magnet.width * scale_val
            elif issubclass(magnet.__class__, Circle):
                magnet.radius = magnet.radius * scale_val
            elif issubclass(magnet.__class__, PolyMagnet):
                magnet.polygon.vertices = (
                    scale_val * _np.array(magnet.polygon.vertices)
                ).tolist()


class Point_Array3(Point_Array2):
    """3D point structure
    This is used to contain three position arrays (x, y, z), and the units
    ('mm', 'cm', etc)
    """

    def __init__(self, x, y, z, unit="mm"):
        """Init Method

        Args:
            x (ndarray): x coordinates
            y (ndarray): y coordinates
            z (ndarray): z coordinates
            unit (str, optional): Unit of length. Defaults to "mm".

        Raises:
            ValueError: Unit must an SI prefix, e.g. km, m, cm, mm
        """
        super().__init__(x, y, unit=unit)
        self.z = _np.asarray(z)

    def __repr__(self) -> str:
        return f"[Unit: {self.unit}\nx: {self.x}\ny: {self.y}\nz: {self.z}]"

    def __str__(self) -> str:
        return f"[Unit: {self.unit}\nx: {self.x}\ny: {self.y}\nz: {self.z}]"

    def change_unit(self, new_unit, get_unit_value=get_unit_value_meter):
        """Converts point array to a different unit. e.g from 'cm' to 'mm'

        Args:
            new_unit (str): unit to be converted to
            get_unit_value (function, optional): Function for checking unit type. Defaults to get_unit_value_meter.
        """
        from ..magnets import Magnet, Prism, Cube, Cylinder, Sphere, Mesh

        current_unit_val = get_unit_value(self.get_unit())
        new_unit_val = get_unit_value(new_unit)
        scale_val = current_unit_val / new_unit_val
        self.x *= scale_val
        self.y *= scale_val
        self.z *= scale_val
        self.unit = new_unit

        for magnet in Magnet.instances:
            magnet.center = scale_val * magnet.center
            if issubclass(magnet.__class__, Prism):
                magnet.a = magnet.a * scale_val
                magnet.b = magnet.b * scale_val
                magnet.c = magnet.c * scale_val
                magnet.width = magnet.width * scale_val
                magnet.height = magnet.height * scale_val
                magnet.depth = magnet.depth * scale_val
            elif issubclass(magnet.__class__, Cube):
                magnet.a = magnet.a * scale_val
                magnet.width = magnet.width * scale_val
            elif issubclass(magnet.__class__, Cylinder):
                magnet.radius = magnet.radius * scale_val
                magnet.length = magnet.length * scale_val
            elif issubclass(magnet.__class__, Sphere):
                magnet.radius = magnet.radius * scale_val
            elif issubclass(magnet.__class__, Mesh):
                magnet.mesh_vectors = magnet.mesh_vectors * scale_val


class Field1(Point_Array1):
    """1D Field vector
    This is used to contain one component (Bz as z), and the units
    ('T', 'mT', etc)
    """

    def __init__(self, z, unit="T"):
        """Init method

        Args:
            z (ndarray): Magnetic field component
            unit (str, optional): Unit of field. Defaults to "T".

        Raises:
            ValueError: Unit must an SI prefix, e.g. T, mT, uT, nT
        """
        super().__init__(z)
        if get_unit_value_tesla(unit) is not None:
            self.unit = unit
        else:
            raise ValueError("Error, not an SI prefix, e.g. T, mT, uT, nT ")

    def __repr__(self) -> str:
        return f"[Unit: {self.unit}\nBz: {self.z}]"

    def __str__(self) -> str:
        return f"[Unit: {self.unit}\nBz: {self.z}]"

    def change_unit(self, new_unit, get_unit_value=get_unit_value_tesla):
        """Converts field array to a different unit. e.g from 'T' to 'mT'

        Args:
            new_unit (str): unit to be converted to
            get_unit_value (function, optional): Function for checking unit type. Defaults to get_unit_value_tesla.
        """
        super().change_unit(new_unit, get_unit_value)


class Field2(Point_Array2):
    """2D Field vector
    This is used to contain two components (x, y), and the units
    ('T', 'mT', etc)
    """

    def __init__(self, x, y, unit="T"):
        """Init method

        Args:
            x (ndarray): Magnetic field component Bx
            y (ndarray): Magnetic field component By
            unit (str, optional): Unit of field. Defaults to "T".

        Raises:
            ValueError: Unit must an SI prefix, e.g. T, mT, uT, nT
        """
        super().__init__(x, y)
        self.n = _np.zeros_like(x)
        if get_unit_value_tesla(unit) is not None:
            self.unit = unit
        else:
            raise ValueError("Error, not an SI prefix, e.g. T, mT, uT, nT ")

    def calc_norm(self):
        """Calculates the norm of the 2D vector"""
        self.n = _np.linalg.norm([self.x, self.y], axis=0)

    def __repr__(self) -> str:
        return f"[Unit: {self.unit}\nBx: {self.x}\nBy: {self.y}\nBn: {self.n}]"

    def __str__(self) -> str:
        return f"[Unit: {self.unit}\nBx: {self.x}\nBy: {self.y}\nBn: {self.n}]"

    def change_unit(self, new_unit, get_unit_value=get_unit_value_tesla):
        """Converts field array to a different unit. e.g from 'T' to 'mT'

        Args:
            new_unit (str): unit to be converted to
            get_unit_value (function, optional): Function for checking unit type. Defaults to get_unit_value_tesla.
        """
        super().change_unit(new_unit, get_unit_value)


class Field3(Point_Array3):
    """2D Field vector
    This is used to contain three components (x, y), and the units
    ('T', 'mT', etc)
    """

    def __init__(self, x, y, z, unit="T"):
        """Init method

        Args:
            x (ndarray): Magnetic field component Bx
            y (ndarray): Magnetic field component By
            z (ndarray): Magnetic field component Bz
            unit (str, optional): Unit of field. Defaults to "T".

        Raises:
            ValueError: Unit must an SI prefix, e.g. T, mT, uT, nT
        """
        super().__init__(x, y, z)
        self.n = _np.zeros_like(x)
        self.unit = unit

    def calc_norm(self):
        """Calculates the norm of the 3D vector"""
        self.n = _np.linalg.norm([self.x, self.y, self.z], axis=0)

    def change_unit(self, new_unit, get_unit_value=get_unit_value_tesla):
        """Converts field array to a different unit. e.g from 'T' to 'mT'

        Args:
            new_unit (str): unit to be converted to
            get_unit_value (function, optional): Function for checking unit type. Defaults to get_unit_value_tesla.
        """
        super().change_unit(new_unit, get_unit_value)

    def __repr__(self) -> str:
        return f"[Unit: {self.unit}\nBx: {self.x}\nBy: {self.y}\nBz: {self.z}\nBn: {self.n}]"

    def __str__(self) -> str:
        return f"[Unit: {self.unit}\nBx: {self.x}\nBy: {self.y}\nBz: {self.z}\nBn: {self.n}]"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# Copyright 2021 Peter Dunne
"""3D Magnet classes

This private module implements the `Prism`, `Cube`, `Cylinder`, and `Sphere`
3D magnet classes. The parent class `Magnet3D` implements the location and orientation
methods, i.e. magnet center and quaternion methods for rotating the magnet with respect
to each principal axis.

TODO:
    * Update __str__ and __repr__ methods to show orientation and magnetisation
"""
import numpy as _np
from ._magnet_base import Magnet
from ..utils._quaternion import Quaternion
from ..utils.global_const import MAG_TOL, PI

from numba import vectorize, float64
from math import sqrt, fabs


__all__ = ["Magnet3D", "Prism", "Cube", "Cylinder", "Sphere"]


class Magnet3D(Magnet):
    """3D Magnet Base Class

    Args:
        Magnet (Magnet): Magnet base parent class

    Returns:
        Magnet3D: 3D magnet object
    """

    mag_type = "Magnet3D"

    def __init__(self, Jr, **kwargs) -> None:
        """Init Method

        Args:
            Jr (float): Signed remnant magnetisation

        Kwargs:
            center (ndarray): Magnet center. Defaults to (0.0, 0.0, 0.0)
            mask_magnet (bool): Flag to mask magnet or not in plots
            alpha (float): Magnet Orientation angle about z (degrees). Defaults to 0.0
            beta (float): Magnet Orientation angle about y (degrees). Defaults to 0.0
            gamma (float): Magnet Orientation angle about x (degrees). Defaults to 0.0
        """
        super().__init__()

        self.Jr = Jr

        self.center = kwargs.pop("center", _np.array([0.0, 0.0, 0.0]))
        self.center = _np.asarray(self.center)

        self._mask_magnet = kwargs.pop("mask_magnet", False)

        # if type(self.center) is tuple:
        #     center = Point3(self.center[0], self.center[1], self.center[2])
        # self.xc = self.center[0]
        # self.yc = self.center[1]
        # self.zc = self.center[2]

        self.alpha = kwargs.pop("alpha", 0.0)  # rotation angle about z
        self.beta = kwargs.pop("beta", 0.0)  # rotation angle about y
        self.gamma = kwargs.pop("gamma", 0.0)  # rotation angle about x

        self.alpha_rad = _np.deg2rad(self.alpha)
        self.beta_rad = _np.deg2rad(self.beta)
        self.gamma_rad = _np.deg2rad(self.gamma)

    def get_center(self):
        """Returns magnet center

        Returns:
            ndarray: [center_x, center_y, center_z]
        """
        return self.center

    def get_Jr(self):
        """Returns local magnetisation orientation

        Must be implemented for all classes
        """
        pass

    def get_orientation(self):
        """Returns magnet orientation, `alpha`, `beta`, `gamma` in degrees

        Returns:
            ndarray: alpha, beta, gamma rotation angles w.r.t z, y, and x axes
        """

        return _np.array([self.alpha, self.beta, self.gamma])

    def _generate_rotation_quaternions(self):

        """Generates single rotation quaternion for all non-zero rotation angles,
        which are:

            alpha: angle in degrees around z-axis
            beta: angle in degrees around y-axis
            gamma: angle in degrees around x-axis

        Returns:
            Quaternion: total rotation quaternion
        """

        # Initialise quaternions
        rotate_about_x = Quaternion()
        rotate_about_y = Quaternion()
        rotate_about_z = Quaternion()

        forward_rotation, reverse_rotation = Quaternion(), Quaternion()

        if _np.fabs(self.alpha_rad) > 1e-4:
            rotate_about_z = Quaternion.q_angle_from_axis(self.alpha_rad, (0, 0, 1))

        if _np.fabs(self.beta_rad) > 1e-4:
            rotate_about_y = Quaternion.q_angle_from_axis(self.beta_rad, (0, 1, 0))

        if _np.fabs(self.gamma_rad) > 1e-4:
            rotate_about_x = Quaternion.q_angle_from_axis(self.gamma_rad, (1, 0, 0))

        # Generate compound rotations
        # Order of rotation: beta  about y, alpha about z, gamma about x
        forward_rotation = rotate_about_x * rotate_about_z * rotate_about_y

        reverse_rotation = forward_rotation.get_conjugate()

        return forward_rotation, reverse_rotation

    def get_field(self, x, y, z):
        """Calculates the magnetic field at point(s) x,y,z due to a 3D magnet
        The calculations are always performed in local coordinates with the centre of the magnet at origin and z magnetisation pointing along the local z' axis.

        The rotations and translations are performed first, and the internal field calculation functions are called.

        Args:
            x (ndarray): x co-ordinates
            y (ndarray): y co-ordinates
            z (ndarray): z co-ordinates

        Returns:
            tuple: Bx(ndarray), By(ndarray), Bz(ndarray) field vector
        """
        from ..utils._routines3D import _tile_arrays, _apply_mask

        # If any rotation angle is set, transform the data
        if _np.any(
            _np.fabs(
                _np.array(
                    [
                        self.alpha_rad,
                        self.beta_rad,
                        self.gamma_rad,
                    ]
                )
            )
            > Magnet.tol
        ):

            forward_rotation, reverse_rotation = self._generate_rotation_quaternions()

            # Generate 3xN array for quaternion rotation
            pos_vec = Quaternion._prepare_vector(
                x - self.center[0], y - self.center[1], z - self.center[2]
            )

            # Rotate points
            x_rot, y_rot, z_rot = forward_rotation * pos_vec

            # Calls internal child method to calculate the field
            B_local = self._get_field_internal(x_rot, y_rot, z_rot)
            mask = self._generate_mask(x_rot, y_rot, z_rot)

            B_local = _apply_mask(self, B_local, mask)

            # Rearrange the field vectors in a 3xN array for quaternion rotation
            Bvec = Quaternion._prepare_vector(B_local.x, B_local.y, B_local.z)

            # Rotate the local fields back into the global frame using quaternions
            Bx, By, Bz = reverse_rotation * Bvec

            # finally return the fields
            return Bx, By, Bz

        else:
            # Otherwise directly calculate the magnetic fields
            B = self._get_field_internal(
                x - self.center[0], y - self.center[1], z - self.center[2]
            )

            xloc, yloc, zloc = _tile_arrays(
                x - self.center[0], y - self.center[1], z - self.center[2]
            )
            mask = self._generate_mask(xloc, yloc, zloc)
            B = _apply_mask(self, B, mask)

            return B.x, B.y, B.z

    def get_force_torque(self):
        """Calculates the force and torque on a magnet due to all other magnets.

        This is a template that needs to be implemented for each magnet.
        """
        pass

    def _get_field_internal(x, y, z):
        """Internal magnetic field calculation method. This should be defined
        for each magnet type.

        Args:
            x (ndarray): x co-ordinates
            y (ndarray): y co-ordinates
            z (ndarray): z co-ordinates
        """
        pass

    def _generate_mask(self, x, y, z):
        """Generates mask of points inside a magnet

        Args:
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates
            z (ndarray): z-coordinates
        """
        pass

    def get_size(self):
        """Returns magnet dimesions

        Must be implemented for each magnet
        """
        pass


class Prism(Magnet3D):
    """Prism 3D Magnet Class

    Args:
        Magnet3D (Magnet3D): 3D magnet parent class

    Returns:
        Prism: Prism magnet object
    """

    mag_type = "Prism"

    def __init__(
        self,
        width=10.0,
        depth=20.0,
        height=30.0,  # magnet dimensions
        Jr=1.0,  # local magnetisation direction
        **kwargs,
    ):
        """Init Method

        Args:
            width (float, optional): Magnet width in x. Defaults to 10.0.
            depth (float, optional): Magnet depth in y. Defaults to 20.0.
            height (float, optional): Magnet height in z. Defaults to 30.0.

        Kwargs:
            center (ndarray): Magnet center. Defaults to (0.0, 0.0, 0.0)
            mask_magnet (bool): Flag to mask magnet or not in plots
            alpha (float): Magnet Orientation angle about z (degrees). Defaults to 0.0
            beta (float): Magnet Orientation angle about y (degrees). Defaults to 0.0
            gamma (float): Magnet Orientation angle about x (degrees). Defaults to 0.0
            phi (float): Angle of magnetisation vector (in degrees) with respect to x-axis. Defaults to 90.0
            theta (float): Angle of magnetisation vector (in degrees) with respect to z-axis. Defaults to 0.0
        """
        self.width = width
        self.depth = depth
        self.height = height

        self.a = width / 2
        self.b = depth / 2
        self.c = height / 2

        super().__init__(Jr, **kwargs)

        self.phi = kwargs.pop("theta", 90.0)
        self.phi_rad = _np.deg2rad(self.phi)
        self.theta = kwargs.pop("phi", 0.0)
        self.theta_rad = _np.deg2rad(self.theta)

        self.Jx = _np.around(
            Jr * _np.cos(self.phi_rad) * _np.sin(self.theta_rad), decimals=6
        )
        self.Jy = _np.around(
            Jr * _np.sin(self.phi_rad) * _np.sin(self.theta_rad), decimals=6
        )
        self.Jz = _np.around(Jr * _np.cos(self.theta_rad), decimals=6)
        self.tol = MAG_TOL  # sufficient for 0.01 degree accuracy

    def __str__(self):
        str = (
            f"{self.__class__.mag_type}\n"
            + f"J: {self.get_Jr()} (T)\n"
            + f"Size: {self.get_size()}\n"
            + f"Center {self.get_center()} \n"
            + f"Orientation alpha,beta,gamma: {self.get_orientation()}\n"
        )
        return str

    def __repr__(self):
        str = (
            f"{self.__class__.mag_type}\n"
            + f"J: {self.get_Jr()} (T)\n"
            + f"Size: {self.get_size()}\n"
            + f"Center {self.get_center()} \n"
            + f"Orientation alpha,beta,gamma: {self.get_orientation()}\n"
        )
        return str

    def get_Jr(self):
        """Returns magnetisation vector J

        Returns:
            ndarray: [Jx, Jy, Jz]
        """
        return _np.array([self.Jx, self.Jy, self.Jz])

    def get_size(self):
        """Returns magnet dimesions

        Returns:
        ndarray: [width, depth, height]
        """
        return _np.array([self.width, self.depth, self.height])

    def get_force_torque(self, num_samples=20, unit="mm"):
        """Calculates the force and torque on a prism magnet due to all other magnets.

        Args:
            num_samples (int, optional): Number of samples per axis per face. Defaults to 20.
            unit (str, optional): Length scale. Defaults to 'mm'.

        Returns:
            tuple: force (ndarray (3,) ) and torque (ndarray (3,) )
        """
        from ..forces._prism_force import calc_force_prism

        force, torque = calc_force_prism(self, num_samples, unit)
        return force, torque

    @staticmethod
    def _F1(a, b, c, x, y, z):
        """Helper Function F1 for 3D prismatic magnets

        Args:
            a (float): magnet half width
            b (float): magnet half depth
            c (float): magnet half height
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates
            z (ndarray): z-coordinates

        Returns:
            ndarray: F1 values
        """
        try:
            # Hide the warning for situtations where there is a divide by zero.
            # This returns a NaN in the array, which is ignored for plotting.
            with _np.errstate(divide="ignore", invalid="ignore"):
                data = _np.arctan(
                    ((y + b) * (z + c))
                    / (
                        (x + a)
                        * _np.sqrt(
                            _np.power((x + a), 2)
                            + _np.power((y + b), 2)
                            + _np.power((z + c), 2)
                        )
                    )
                )
        except ValueError:
            data = _np.NaN
        return data

    @staticmethod
    def _F2(a, b, c, x, y, z):
        """Helper Function F2 for 3D prismatic magnets

        Args:
            a (float): magnet half width
            b (float): magnet half depth
            c (float): magnet half height
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates
            z (ndarray): z-coordinates

        Returns:
            ndarray: F2 values
        """
        try:
            xa_sq = _np.power((x + a), 2)
            yb_sq = _np.power((y + b), 2)
            zc_sq = _np.power((z + c), 2)
            znc_sq = _np.power((z - c), 2)
            # Hide the warning for situtations where there is a divide by zero.
            # This returns a NaN in the array, which is ignored for plotting.
            with _np.errstate(divide="ignore", invalid="ignore"):
                data = (_np.sqrt(xa_sq + yb_sq + znc_sq) + c - z) / (
                    _np.sqrt(xa_sq + yb_sq + zc_sq) - c - z
                )
        except ValueError:
            data = _np.NaN
        return data

    def _get_field_internal(self, x, y, z):
        """Internal magnetic field calculation methods.
        Iterates over each component of the prism magnetised in x, y, and z
        (in local coordinates).

        Args:
            x (ndarray): x co-ordinates
            y (ndarray): y co-ordinates
            z (ndarray): z co-ordinates

        Returns:
            Field3: Magnetic field array structure
        """
        from ..utils._routines3D import _allocate_field_array3

        B = _allocate_field_array3(x, y, z)
        # Magnetic field due to component of M magnetised in x
        if _np.fabs(self.Jx) > self.tol:
            Bx, By, Bz = self._calcB_prism_x(x, y, z)
            B.x += Bx
            B.y += By
            B.z += Bz

        # Magnetic field due to component of M magnetised in y
        if _np.fabs(self.Jy) > self.tol:
            Bx, By, Bz = self._calcB_prism_y(x, y, z)
            B.x += Bx
            B.y += By
            B.z += Bz

        # Magnetic field due to component of M magnetised in z
        if _np.fabs(self.Jz) > self.tol:
            Bx, By, Bz = self._calcB_prism_z(x, y, z)
            B.x += Bx
            B.y += By
            B.z += Bz
        return B

    def _calcBx_prism_x(self, a, b, c, Jr, x, y, z):
        """Calculates x component of magnetic field for prism magnet
        magnetised in x

        Args:
            Args:
            a (float): magnet half width
            b (float): magnet half depth
            c (float): magnet half height
            Jr (float): Remnant magnetisation
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates
            z (ndarray): z-coordinates

        Returns:
            ndarray: Bx magnetic field component
        """

        try:
            data = -(Jr / (4 * PI)) * (
                self._F1(a, b, c, -x, y, z)
                + self._F1(a, b, c, -x, y, -z)
                + self._F1(a, b, c, -x, -y, z)
                + self._F1(a, b, c, -x, -y, -z)
                + self._F1(a, b, c, x, y, z)
                + self._F1(a, b, c, x, y, -z)
                + self._F1(a, b, c, x, -y, z)
                + self._F1(a, b, c, x, -y, -z)
            )
        except ValueError:
            data = _np.NaN
        return data

    def _calcBy_prism_x(self, a, b, c, Jr, x, y, z):
        """Calculates y component of magnetic field for prism magnet
        magnetised in x

        Args:
            a (float): magnet half width
            b (float): magnet half depth
            c (float): magnet half height
            Jr (float): Remnant magnetisation
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates
            z (ndarray): z-coordinates

        Returns:
            ndarray: By magnetic field component
        """

        try:
            data = _np.log(
                self._F2(a, b, c, -x, -y, z)
                * self._F2(a, b, c, x, y, z)
                / (self._F2(a, b, c, -x, y, z) * self._F2(a, b, c, x, -y, z))
            )
            data *= Jr / (4 * PI)
        except ValueError:
            data = _np.NaN
        return data

    def _calcBz_prism_x(self, a, b, c, Jr, x, y, z):
        """Calculates z component of magnetic field for prism magnet
        magnetised in x

        Args:
            a (float): magnet half width
            b (float): magnet half depth
            c (float): magnet half height
            Jr (float): Remnant magnetisation
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates
            z (ndarray): z-coordinates

        Returns:
            ndarray: Bz magnetic field component
        """
        try:
            # Hide the warning for situtations where there is a divide by zero.
            # This returns a NaN in the array, which is ignored for plotting.
            with _np.errstate(divide="ignore", invalid="ignore"):
                data = _np.log(
                    self._F2(a, c, b, -x, -z, y)
                    * self._F2(a, c, b, x, z, y)
                    / (self._F2(a, c, b, -x, z, y) * self._F2(a, c, b, x, -z, y))
                )
            data *= Jr / (4 * PI)
        except ValueError:
            data = _np.NaN
        return data

    def _calcB_prism_x(self, x, y, z):
        """Calculates agnetic field vector due to magnet magnetised in x

        Args:
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates
            z (ndarray): z-coordinates

        Returns:
            tuple: Bx (ndarray), By (ndarray), Bz (ndarray)
        """

        a = self.a
        b = self.b
        c = self.c
        Jr = self.Jr

        Bx = self._calcBx_prism_x(a, b, c, Jr, x, y, z)
        By = self._calcBy_prism_x(a, b, c, Jr, x, y, z)
        Bz = self._calcBz_prism_x(a, b, c, Jr, x, y, z)
        return Bx, By, Bz

    def _calcB_prism_z(self, x, y, z):
        """Calculates agnetic field vector due to magnet magnetised in y

        Args:
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates
            z (ndarray): z-coordinates

        Returns:
            tuple: Bx (ndarray), By (ndarray), Bz (ndarray)
        """
        a = self.a
        b = self.b
        c = self.c
        Jr = self.Jr

        Bx = self._calcBz_prism_x(-c, b, a, Jr, -z, y, x)
        By = self._calcBy_prism_x(-c, b, a, Jr, -z, y, x)
        Bz = -1 * self._calcBx_prism_x(-c, b, a, Jr, -z, y, x)
        return Bx, By, Bz

    def _calcB_prism_y(self, x, y, z):
        """Calculates agnetic field vector due to magnet magnetised in z

        Args:
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates
            z (ndarray): z-coordinates

        Returns:
            tuple: Bx (ndarray), By (ndarray), Bz (ndarray)
        """
        a = self.a
        b = self.b
        c = self.c
        Jr = self.Jr

        Bx = self._calcBy_prism_x(-b, a, c, Jr, -y, x, z)
        By = -1 * self._calcBx_prism_x(-b, a, c, Jr, -y, x, z)
        Bz = self._calcBz_prism_x(-b, a, c, Jr, -y, x, z)
        return Bx, By, Bz

    def _generate_mask(self, x, y, z):
        """Generates mask of points inside a magnet

        Args:
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates
            z (ndarray): z-coordinates
        """
        w, d, h = self.get_size()
        xn = -w / 2
        xp = w / 2
        yn = -d / 2
        yp = d / 2
        zn = -h / 2
        zp = h / 2

        mask_x = _np.logical_and(x > xn, x < xp)
        mask_y = _np.logical_and(y > yn, y < yp)
        mask_z = _np.logical_and(z > zn, z < zp)

        # merge logical masks
        mask = _np.logical_and(mask_x, mask_z)
        mask = _np.logical_and(mask, mask_y)

        return mask


class Cube(Prism):
    """Cube 3D Magnet Class

    Args:
        Prism (Prism): Prism magnet parent class
    """

    mag_type = "Cube"

    def __init__(
        self,
        width=10.0,  # magnet dimensions
        Jr=1.0,  # local magnetisation direction
        **kwargs,
    ):
        """Init method

        Args:
            width (float, optional): Cube side length. Defaults to 10.0.

        Kwargs:
            center (ndarray): Magnet center. Defaults to (0.0, 0.0, 0.0)
            mask_magnet (bool): Flag to mask magnet or not in plots
            alpha (float): Magnet Orientation angle about z (degrees). Defaults to 0.0
            beta (float): Magnet Orientation angle about y (degrees). Defaults to 0.0
            gamma (float): Magnet Orientation angle about x (degrees). Defaults to 0.0
            phi (float): Angle of magnetisation vector (in degrees) with respect to x-axis. Defaults to 90.0
            theta (float): Angle of magnetisation vector (in degrees) with respect to z-axis. Defaults to 0.0
        """

        super().__init__(width=width, depth=width, height=width, Jr=Jr, **kwargs)


class Cylinder(Magnet3D):
    """Cylinder 3D Magnet Class

    Args:
        Magnet3D (Magnet3D): 3D magnet parent class

    Returns:
        Cylinder: Cylinder 3D magnet object
    """

    mag_type = "Cylinder"

    def __init__(
        self,
        radius=10.0,
        length=10.0,  # magnet dimensions
        Jr=1.0,  # local magnetisation direction
        **kwargs,
    ):
        """Init Method

        Args:
            radius (float, optional): radius. Defaults to 10.0.
            length (float, optional): length. Defaults to 10.0.

        Kwargs:
            center (ndarray): Magnet center. Defaults to (0.0, 0.0, 0.0)
            mask_magnet (bool): Flag to mask magnet or not in plots
            alpha (float): Magnet Orientation angle about z (degrees). Defaults to 0.0
            beta (float): Magnet Orientation angle about y (degrees). Defaults to 0.0
            gamma (float): Magnet Orientation angle about x (degrees). Defaults to 0.0
            phi (float): Angle of magnetisation vector (in degrees) with respect to x-axis. Defaults to 90.0
            theta (float): Angle of magnetisation vector (in degrees) with respect to z-axis. Defaults to 0.0
        """
        super().__init__(Jr, **kwargs)
        self.radius = radius
        self.length = length

        self.center = kwargs.pop("center", _np.array([0.0, 0.0, 0.0]))
        self.center = _np.asarray(self.center)

    def __str__(self):
        str = (
            f"{self.__class__.mag_type}\n"
            + f"J: {self.get_Jr()} (T)\n"
            + f"Size: {self.get_size()}\n"
            + f"Center {self.get_center()}\n"
            + f"Orientation alpha,beta,gamma: {self.get_orientation()}\n"
        )
        return str

    def __repr__(self):
        str = (
            f"{self.__class__.mag_type}\n"
            + f"J: {self.Jr} (T)\n"
            + f"Size: {self.get_size() }\n"
            + f"Center {self.get_center()}\n"
            + f"Orientation alpha,beta,gamma: {self.get_orientation()}\n"
        )
        return str

    def get_size(self):
        """Returns magnet dimesions

        Returns:
            size[ndarray]: numpy array [radius, length]
        """
        return _np.array([self.radius, self.length])

    def get_Jr(self):
        return _np.array([0.0, 0.0, self.Jr])

    def get_force_torque(self, num_samples=20, unit="mm"):
        """Calculates the force and torque on a cylinder magnet due to all other magnets.

        Args:
            num_samples (int, optional): Number of samples per axis per face. Defaults to 20.
            unit (str, optional): Length scale. Defaults to 'mm'.

        Returns:
            tuple: force (ndarray (3,) ) and torque (ndarray (3,) )
        """
        from ..forces._cylinder_force import calc_force_cylinder

        force, torque = calc_force_cylinder(self, num_samples, unit)
        return force, torque

    @staticmethod
    @vectorize([float64(float64, float64, float64, float64)], target="parallel")
    def _cel(kc, p, c, s):
        """Bulirsch's complete elliptic integral
        See NIST Handbook of Mathematical Functions, http://dlmf.nist.gov/19.2

        Numba is used to create a compiled Numpy ufunc that accepts numpy arrays.

        Args:
            kc (float/ndarray): elliptical modulus
            p (float/ndarray): real parameter
            c (float/ndarray): real parameter
            s (float/ndarray): real parameter

        Returns:
            float/ndarray: result of computing complete elliptic integral
        """
        if kc == 0:
            data = _np.NaN
            return data
        else:
            errtol = 0.000001
            k = fabs(kc)
            pp = p
            cc = c
            ss = s
            em = 1.0

            if p > 0:
                pp = sqrt(p)
                ss = s / pp
            else:
                f = kc * kc
                q = 1.0 - f
                g = 1.0 - pp
                f = f - pp
                q = q * (ss - c * pp)
                pp = sqrt(f / g)
                cc = (c - ss) / g
                ss = -q / (g * g * pp) + cc * pp
            f = cc
            cc = cc + ss / pp
            g = k / pp
            ss = 2 * (ss + f * g)
            pp = g + pp
            g = em
            em = k + em
            kk = k

            while fabs(g - k) > g * errtol:
                k = 2 * sqrt(kk)
                kk = k * em
                f = cc
                cc = cc + ss / pp
                g = kk / pp
                ss = 2 * (ss + f * g)
                pp = g + pp
                g = em
                em = k + em
            data = (PI / 2.0) * (ss + cc * em) / (em * (em + pp))
            return data

    def _get_field_internal(self, x, y, z):
        """Internal magnetic field calculation methods.
        Calculates the field due to a cylindrical magnet/solenoid magnetised along z
        (in local coordinates).

        Args:
            x (array): x co-ordinates
            y (array): y co-ordinates
            z (array): z co-ordinates

        Returns:
            Field: Magnetic field array
        """
        from ..utils._conversions import cart2pol, pol2cart
        from ..utils._routines3D import _allocate_field_array3

        B = _allocate_field_array3(x, y, z)

        # Convert cartesian coordinates to cylindrical
        rho, phi = cart2pol(x, y)

        Brho, B.z = self._calcB_cyl(rho, z)

        # Convert magnetic fields from cylindrical to cartesian
        # We use pol2cart because Bphi is zero
        # If Bphi was != 0, then would have to use
        # `..utils._conversions.vector_pol2cart(Brho, Bphi, phi)`

        B.x, B.y = pol2cart(Brho, phi)

        return B

    def _calcB_cyl(self, rho, z):
        """Calculates the magnetic field due to a solenoid/cylinder in
        polar cylindrical coordinates

        Args:
            rho (array): radial coordinates
            z (array): axial coordinates

        Returns:
            tuple: Brho (ndarray), Bz (ndarray) magnetic field components
        """
        a = self.radius
        b = self.length / 2
        B0 = self.Jr / PI

        zp = z + b
        zn = z - b

        zp_sq = _np.power(zp, 2)
        zn_sq = _np.power(zn, 2)
        rho_a_sq = _np.power(rho + a, 2)
        nrho_a_sq = _np.power(a - rho, 2)

        alphap = a / _np.sqrt(zp_sq + rho_a_sq)
        alphan = a / _np.sqrt(zn_sq + rho_a_sq)

        betap = zp / _np.sqrt(zp_sq + rho_a_sq)
        betan = zn / _np.sqrt(zn_sq + rho_a_sq)

        gamma = (a - rho) / (a + rho)

        kp = _np.sqrt((zp_sq + nrho_a_sq) / (zp_sq + rho_a_sq))

        kn = _np.sqrt((zn_sq + nrho_a_sq) / (zn_sq + rho_a_sq))

        Brho = B0 * (
            alphap * self._cel(kp, 1, 1, -1) - alphan * self._cel(kn, 1, 1, -1)
        )

        Bz = (B0 * a / (a + rho)) * (
            betap * self._cel(kp, gamma ** 2, 1, gamma)
            - betan * self._cel(kn, gamma ** 2, 1, gamma)
        )
        return Brho, Bz

    def _generate_mask(self, x, y, z):
        """Generates mask of points inside a cylindrical magnet

        Args:
            x (ndarray): x-coordinates
            y (ndarray): y-coordinates
            z (ndarray): z-coordinates
        """

        radius, length = self.get_size()
        data_norm = x ** 2 + y ** 2
        zn = -length / 2
        zp = length / 2

        mask_rho = data_norm < radius ** 2
        mask_z = _np.logical_and(z > zn, z < zp)
        mask = _np.logical_and(mask_rho, mask_z)

        return mask


class Sphere(Magnet3D):
    """Sphere 3D Magnet Class

    Args:
        Magnet3D (Magnet3D): 3D magnet parent class

    Returns:
        Sphere: Sphere 3D magnet object
    """

    mag_type = "Sphere"

    def __init__(
        self,
        radius=10.0,
        Jr=1.0,  # local magnetisation direction
        **kwargs,
    ):
        """Init Method

        Args:
            radius (float, optional): radius. Defaults to 10.0.
            Jr (float, optional): remnant magnetisation. Defaults to 1.0.

        Kwargs:
            center (ndarray): Magnet center. Defaults to (0.0, 0.0, 0.0)
            mask_magnet (bool): Flag to mask magnet or not in plots
            alpha (float): Magnet Orientation angle about z (degrees). Defaults to 0.0
            beta (float): Magnet Orientation angle about y (degrees). Defaults to 0.0
            gamma (float): Magnet Orientation angle about x (degrees). Defaults to 0.0
        """
        super().__init__(Jr, **kwargs)
        self.radius = radius

        self.phi = kwargs.pop("phi", None)
        self.theta = kwargs.pop("theta", None)

        if self.phi is not None or self.theta is not None:
            print("Warning, the magnetisation of a sphere is always in z.")
            print("Do not use phi or theta.")
            print("To rotate the magnetisation, use alpha, beta and gamma")

        self.center = kwargs.pop("center", _np.array([0.0, 0.0, 0.0]))
        self.center = _np.asarray(self.center)

    def __str__(self):
        str = (
            f"{self.__class__.mag_type}\n"
            + f"J: {self.get_Jr()} (T)\n"
            + f"Size: {self.get_size()}\n"
            + f"Center {self.get_center()}\n"
            + f"Orientation alpha,beta,gamma: {self.get_orientation()}\n"
        )
        return str

    def __repr__(self):
        str = (
            f"{self.__class__.mag_type}\n"
            + f"J: {self.get_Jr()} (T)\n"
            + f"Size: {self.get_size() }\n"
            + f"Center {self.get_center()}\n"
            + f"Orientation alpha,beta,gamma: {self.get_orientation()}\n"
        )
        return str

    def get_size(self):
        """Returns magnet dimesions

        Returns:
            size[ndarray]: numpy array [radius, length]
        """
        return _np.array([self.radius])

    def get_Jr(self):
        return _np.array([0, 0, self.Jr])

    def get_force_torque(self, num_samples=100, unit="mm"):
        """Calculates the force and torque on a sphere magnet due to all other magnets.

        Args:
            num_samples (int, optional): Number of samples per axis per face. Defaults to 100.
            unit (str, optional): Length scale. Defaults to 'mm'.

        Returns:
            tuple: force (ndarray (3,) ) and torque (ndarray (3,) )
        """
        from ..forces._sphere_force import calc_force_sphere

        force, torque = calc_force_sphere(self, num_samples, unit)
        return force, torque

    def _get_field_internal(self, x, y, z):
        """Internal magnetic field calculation methods.
        Calculates the field due to a spherical magnet magnetised along z
        (in local coordinates). Returns a dipolar field outside the magnet

        Args:
            x (float/array): x co-ordinates
            y (float/array): y co-ordinates
            z (float/array): z co-ordinates

        Returns:
            Vector3: Magnetic field array
        """
        from ..utils._conversions import cart2sph, sphere_sph2cart
        from ..utils._routines3D import _allocate_field_array3

        B = _allocate_field_array3(x, y, z)

        # Convert to spherical coordinates
        r, theta, phi = cart2sph(x, y, z)

        # Calculates field for sphere magnetised along z
        Br, Btheta = self._calcB_spherical(r, theta)

        # Convert magnetic fields from spherical to cartesian
        B.x, B.y, B.z = sphere_sph2cart(Br, Btheta, theta, phi)
        return B

    def _calcB_spherical(self, r, theta):
        """Calculates the magnetic field due to due to a sphere at any point
        in spherical coordinates

        Args:
            r (float/array): radial coordinates
            theta (float/array): azimuthal coordinates

        Returns:
            tuple: Br, Btheta
        """

        # Hide the warning for situtations where there is a divide by zero.
        # This returns a NaN in the array, which is ignored for plotting.
        with _np.errstate(divide="ignore", invalid="ignore"):
            preFac = self.Jr * (self.radius ** 3 / r ** 3) / 3.0

        Br = preFac * 2.0 * _np.cos(theta)
        Btheta = preFac * _np.sin(theta)

        return Br, Btheta

    def _generate_mask(self, x, y, z):
        """Generates mask of points inside a spherical magnet

        Args:
            x (ndarray/float): x-coordinates
            y (ndarray/float): y-coordinates
            z (ndarray/float): z-coordinates
        """

        data_norm = x ** 2 + y ** 2 + z ** 2
        mask = data_norm < self.radius ** 2

        return mask
